from pytest import approx
from pathlib import Path
import numpy as np
import copy
import os

import openmdao.api as om

from hopp.simulation import HoppInterface
from hopp.utilities import load_yaml

from greenheart.tools.optimization.openmdao import GreenHeartComponent, HOPPComponent, TurbineDistanceComponent, BoundaryDistanceComponent
from greenheart.tools.optimization.gc_run_greenheart import run_greenheart
from greenheart.simulation.greenheart_simulation import GreenHeartSimulationConfig

import unittest

ROOT_DIR = Path(__file__).parents[2]

solar_resource_file = ROOT_DIR / "resource_files" / "solar" / "35.2018863_-101.945027_psmv3_60_2012.csv"
wind_resource_file = ROOT_DIR / "resource_files" / "wind" / "35.2018863_-101.945027_windtoolkit_2012_60min_80m_100m.srw"
floris_input_file = Path(__file__).absolute().parent / "inputs" / "floris_input.yaml"
hopp_config_filename = Path(__file__).absolute().parent / "inputs" / "hopp_config.yaml"
hopp_config_steel_ammonia_filename = Path(__file__).absolute().parent / "input_files" / "plant" / "hopp_config.yaml"
greenheart_config_onshore_filename = Path(__file__).absolute().parent / "input_files" / "plant" / "greenheart_config_onshore.yaml"
turbine_config_filename = Path(__file__).absolute().parent / "input_files" / "turbines" / "osw_18MW.yaml"
floris_input_filename_steel_ammonia = Path(__file__).absolute().parent / "input_files" / "floris" / "floris_input_osw_18MW.yaml"


from ORBIT.core.library import initialize_library

dirname = os.path.dirname(__file__)
orbit_library_path = os.path.join(dirname, "input_files/")

initialize_library(orbit_library_path)
offshore_hopp_config_wind_wave_solar_battery = os.path.join(
        orbit_library_path, f"plant/hopp_config_wind_wave_solar_battery.yaml"
)
offshore_greenheart_config = os.path.join(
    orbit_library_path, f"plant/greenheart_config.yaml"
)
offshore_turbine_model = "osw_18MW"
offshore_turbine_config = os.path.join(
    orbit_library_path, f"turbines/{offshore_turbine_model}.yaml"
)
offshore_floris_config = os.path.join(
    orbit_library_path, f"floris/floris_input_{offshore_turbine_model}.yaml"
)
offshore_orbit_config = os.path.join(
    orbit_library_path, f"plant/orbit-config-{offshore_turbine_model}-stripped.yaml"
)

rtol = 1E-5

def setup_hopp():
    turbine_x = np.array([2.0, 4.0, 6.0, 8.0, 10.0, 12.0])*100.0
    turbine_y = np.array([2.0, 2.0, 4.0, 4.0, 8.0, 8.0])*100.0

    hybrid_config_dict = load_yaml(hopp_config_filename)

    hybrid_config_dict["site"]["solar_resource_file"] = solar_resource_file
    hybrid_config_dict["site"]["solar"] = "true"

    hybrid_config_dict["site"]["wind_resource_file"] = wind_resource_file
    hybrid_config_dict["technologies"]["wind"]["floris_config"] = floris_input_file


    hybrid_config_dict["site"]["desired_schedule"] = [80000.0]*8760
    hybrid_config_dict["technologies"]["battery"] = {"system_capacity_kwh": 400,
                                                            "system_capacity_kw": 100,
                                                            "minimum_SOC": 20.0,
                                                            "maximum_SOC": 100.0,
                                                            "initial_SOC": 90.0}
    hybrid_config_dict["config"]["dispatch_options"] = {"battery_dispatch": "load_following_heuristic",
                                                            "solver": "cbc",
                                                            "n_look_ahead_periods": 48,
                                                            "grid_charging": False,
                                                            "pv_charging_only": False,
                                                            "include_lifecycle_count": False}

    design_variables = ["pv_capacity_kw", "turbine_x"]

    technologies = hybrid_config_dict["technologies"]
    solar_wind_hybrid = {key: technologies[key] for key in ('pv', 'wind', 'battery', 'grid')}
    hybrid_config_dict["technologies"] = solar_wind_hybrid
    hi = HoppInterface(hybrid_config_dict)

    model = om.Group()

    model.add_subsystem('hopp', HOPPComponent(hi=hi, verbose=False, turbine_x_init=turbine_x, turbine_y_init=turbine_y, design_variables=design_variables), promotes=["*"])

    prob = om.Problem(model)
    prob.setup()

    prob.run_model()

    return prob, turbine_x, hybrid_config_dict

def setup_greenheart():
    config = GreenHeartSimulationConfig(
    filename_hopp_config=hopp_config_steel_ammonia_filename,
    filename_greenheart_config=greenheart_config_onshore_filename,
    filename_turbine_config=turbine_config_filename,
    filename_floris_config=floris_input_filename_steel_ammonia,
    verbose=False,
    show_plots=False,
    save_plots=False,
    output_dir= str(Path(__file__).absolute().parent / "output"),
    use_profast=True,
    post_processing=False,
    incentive_option=1,
    plant_design_scenario=9,
    output_level=7,
    )
    
    # based on 2023 ATB moderate case for onshore wind
    config.hopp_config["config"]["cost_info"]["wind_installed_cost_mw"] = 1434000.0 
    # based on 2023 ATB moderate case for onshore wind
    config.hopp_config["config"]["cost_info"]["wind_om_per_kw"] = 29.567
    config.hopp_config["technologies"]["wind"]["fin_model"]["system_costs"]["om_fixed"][0] = config.hopp_config["config"]["cost_info"]["wind_om_per_kw"]
    # set skip_financial to false for onshore wind
    config.hopp_config["config"]["simulation_options"]["wind"]["skip_financial"] = False

    
    config.greenheart_config["opt_options"] = {
            "opt_flag": True,
            "general": {
                "folder_output": "output",
                "fname_output": "test_run_greenheart_optimization",
            },
            "design_variables": {
                "electrolyzer_rating_kw": {
                    "flag": True,
                    "lower": 10000.0,
                    "upper": 200000.0,
                    "units": "kW",
                },
                "pv_capacity_kw": {
                    "flag": False,
                    "lower": 1000.0,
                    "upper": 1500000.0,
                    "units": "kW",
                },
                "wave_capacity_kw": {
                    "flag": False,
                    "lower": 1000.0,
                    "upper": 1500000.0,
                    "units": "kW",
                },
                "battery_capacity_kw": {
                    "flag": False,
                    "lower": 1000.0,
                    "upper": 1500000.0,
                    "units": "kW",
                },
                "battery_capacity_kwh": {
                    "flag": False,
                    "lower": 1000.0,
                    "upper": 1500000.0,
                    "units": "kW*h",
                },
                "battery_capacity_kwh": {
                    "flag": False,
                    "lower": 1000.0,
                    "upper": 1500000.0,
                    "units": "kW*h",
                },
                "turbine_x": {
                    "flag": False,
                    "lower": 0.0,
                    "upper": 1500000.0,
                    "units": "m",
                },
                "turbine_y": {
                    "flag": False,
                    "lower": 0.0,
                    "upper": 1500000.0,
                    "units": "m",
                },
            },
            "constraints": {
                "turbine_spacing": {
                    "flag": False,
                    "lower": 0.0,
                },
                "boundary_distance": {
                    "flag": False,
                    "lower": 0.0,
                },
            "pv_to_platform_area_ratio": {
                "flag": False, 
                "upper": 1.0, # relative size of solar pv area to platform area
                },
            "user": {}
            },
            "merit_figure": "lcoh",
            "merit_figure_user": {
                "name": "lcoh",
                "max_flag": False,
                "ref": 1.0, # value of objective that scales to 1.0
            },
            "driver": {
                "optimization": {
                    "flag": True,
                    "solver": "SLSQP",
                    "tol": 1E-6,
                    "max_major_iter": 1,
                    "max_minor_iter": 2,
                    "gradient_method": "openmdao",
                    # "time_limit": 10, # (sec) optional
                    # "hist_file_name": "snopt_history.txt", # optional
                    "verify_level": -1, # optional
                    "step_calc": None,
                    "form": "forward", # type of finite differences to use, can be one of ["forward", "backward", "central"]
                    "debug_print": False,
                },
                "design_of_experiments": {
                    "flag": False,
                    "run_parallel": False,
                    "generator": "FullFact", # [Uniform, FullFact, PlackettBurman, BoxBehnken, LatinHypercube]
                    "num_samples": 1, # Number of samples to evaluate model at (Uniform and LatinHypercube only)
                    "seed": 2,
                    "levels":  50, #  Number of evenly spaced levels between each design variable lower and upper bound (FullFactorial only)
                    "criterion": None, # [None, center, c, maximin, m, centermaximin, cm, correelation, corr]
                    "iterations": 1,
                    "debug_print": False
                },
                "step_size_study": {
                    "flag": False  
                },
            },
            "recorder": {
                "flag": True,
                "file_name": str(Path(__file__).absolute().parent / "output" / "recorder.sql"),
                "includes": False,
            },
        }

    return config
    

def test_boundary_distance_component(subtests):
    turbine_x = np.array([2.0, 4.0, 6.0, 2.0, 4.0, 6.0])*100.0
    turbine_y = np.array([2.0, 2.0, 2.0, 4.0, 4.0, 4.0])*100.0
    
    config_dict = load_yaml(hopp_config_filename)
    config_dict["site"]["wind_resource_file"] = wind_resource_file
    config_dict["technologies"]["wind"]["floris_config"] = floris_input_file
    hi = HoppInterface(config_dict)

    model = om.Group()

    model.add_subsystem('boundary_constraint', BoundaryDistanceComponent(hopp_interface=hi, turbine_x_init=turbine_x, turbine_y_init=turbine_y), promotes=["*"])

    prob = om.Problem(model)
    prob.setup()

    prob.run_model()
    
    with subtests.test("test distance inside"):
        assert prob["boundary_distance_vec"][0] == 200.0
    
    with subtests.test("test_distance_outside"):

        prob.set_val('boundary_constraint.turbine_x', np.array([-2.0, 4.0, 6.0, 2.0, 4.0, 6.0])*100.0)
        prob.set_val('boundary_constraint.turbine_y', np.array([2.0, 2.0, 2.0, 4.0, 4.0, 4.0])*100.0)
        prob.run_model()
        
        assert prob["boundary_distance_vec"][0] == -200.0

    # TODO add analytic derivatives and test

def test_turbine_distance_component(subtests):

    turbine_x = np.array([2.0, 4.0, 6.0])*100.0
    turbine_y = np.array([2.0, 2.0, 4.0])*100.0

    model = om.Group()

    model.add_subsystem('boundary_constraint', TurbineDistanceComponent(turbine_x_init=turbine_x, turbine_y_init=turbine_y), promotes=["*"])

    prob = om.Problem(model)
    prob.setup()

    prob.run_model()
    
    expected_distances = np.array([200.0, np.sqrt(400**2 + 200**2), np.sqrt(2*200**2)])
    for i in range(len(turbine_x)):
        with subtests.test(f"for element {i}"):
            assert prob["spacing_vec"][i] == expected_distances[i]

def test_hopp_component(subtests):
    
    prob, turbine_x, hybrid_config_dict = setup_hopp()
    
    with subtests.test("inputs_turbine_x"):
        assert prob.get_val('turbine_x')[0] == approx(turbine_x[0])

    with subtests.test("inputs_pv_capacity_kw"):
        assert prob.get_val('pv_capacity_kw')[0] == approx(hybrid_config_dict["technologies"]["pv"]["system_capacity_kw"])

    with subtests.test("costs_pv_capex"):
        assert prob.get_val('pv_capex')[0] == approx(14400000.0)

    # with subtests.test("costs_pv_opex"):
    #     assert prob.get_val('pv_opex')[0] == approx(0.0)

    with subtests.test("costs_wind_capex"):
        assert prob.get_val('wind_capex')[0] == approx(43620000.0)

    # with subtests.test("cost_wind_opex"):
    #     assert prob.get_val('wind_opex')[0] == approx(0.0)

    with subtests.test("costs_battery_capex"):
        assert prob.get_val('battery_capex')[0] == approx(163100.0)

    # with subtests.test("costs_battery_opex"):
    #     assert prob.get_val('battery_opex')[0] == approx(0.0)

    with subtests.test("costs_hybrid_electrical_generation_capex"):
        assert prob.get_val('hybrid_electrical_generation_capex')[0] == approx(58183100.0)

    with subtests.test("costs_total_capex_equals_sum"):
        assert prob.get_val('hybrid_electrical_generation_capex')[0] == approx(14400000.0 + 43620000.0 + 163100.0)
        
    # with subtests.test("costs_hybrid_electrical_generation_opex"):
    #     assert prob.get_val('hybrid_electrical_generation_opex')[0] == approx(0.0)

    with subtests.test("changes_turbine_x"):
        new_x = copy.deepcopy(turbine_x)
        new_x[0] = 0.0
        prob.set_val("turbine_x", new_x)
        assert prob.get_val('turbine_x')[0] == approx(new_x[0])

    with subtests.test("changes_pv_capacity_kw_new"):
        new_pv_capacity_kw = 50
        prob.set_val("pv_capacity_kw", new_pv_capacity_kw)
        assert prob.get_val('pv_capacity_kw')[0] == new_pv_capacity_kw


def test_greenheart_component(subtests):
    
    config = setup_greenheart()

    model = om.Group()

    model.add_subsystem('greenheart', GreenHeartComponent(config=config, design_variables=["electrolyzer_rating_kw"]),  promotes=["*"])

    prob = om.Problem(model)
    prob.setup()
    prob.run_model()

    # TODO base this test value on something
    with subtests.test("lcoh"):
        assert prob["lcoh"][0] == approx(3.1691092704830357, rel=rtol)

    # TODO base this test value on something
    with subtests.test("lcoe"):
        assert prob["lcoe"][0] == approx(0.03486193, rel=rtol)

    # TODO base this test value on something
    with subtests.test("steel_finance"):
        lcos_expected = 1357.046163641118
        assert prob["lcos"][0]  == approx(lcos_expected, rel=rtol)

    # TODO base this test value on something
    with subtests.test("ammonia_finance"):
        lcoa_expected = 1.04190962
        assert prob["lcoa"][0]  == approx(lcoa_expected, rel=rtol)

def test_run_greenheart_run_only(subtests):
    config = setup_greenheart()
    prob, config = run_greenheart(config, run_only=True)
    
    # TODO base this test value on something
    with subtests.test("lcoh"):
        assert prob["lcoh"][0] == approx(3.1691092704830357, rel=rtol)

    # TODO base this test value on something
    with subtests.test("lcoe"):
        assert prob["lcoe"] == approx(0.03486193, rel=rtol)

    # TODO base this test value on something
    with subtests.test("steel_finance"):
        lcos_expected = 1357.046163641118
        assert prob["lcos"]  == approx(lcos_expected, rel=rtol)

    # TODO base this test value on something
    with subtests.test("ammonia_finance"):
        lcoa_expected = 1.04190962
        assert prob["lcoa"]  == approx(lcoa_expected, rel=rtol)

def test_run_greenheart_optimize(subtests):
    
    config = setup_greenheart()

    prob, config = run_greenheart(config, run_only=False)

    cr = om.CaseReader(Path(__file__).absolute().parent / "output" / "recorder.sql")

    # get initial LCOH
    case = cr.get_case(0)
    lcoh_init = case.get_val("lcoh", units='USD/kg')[0]

    # get final LCOH
    case = cr.get_case(-1)
    lcoh_final = case.get_val("lcoh", units='USD/kg')[0]
    
    with subtests.test("lcoh"):
        assert lcoh_final < lcoh_init








import os

from pytest import approx, warns, raises
import yaml
import warnings
import pathlib

from greenheart.simulation.greenheart_simulation import (
    run_simulation,
    GreenHeartSimulationConfig,
)

from greenheart.tools.optimization.gc_run_greenheart import run_greenheart

from hopp.utilities.keys import set_nrel_key_dot_env
from greenheart.tools.eco.utilities import visualize_plant, ceildiv

set_nrel_key_dot_env()

from ORBIT.core.library import initialize_library

dirname = os.path.dirname(__file__)
orbit_library_path = os.path.join(dirname, "input_files/")

initialize_library(orbit_library_path)

turbine_model = "osw_18MW"
filename_turbine_config = os.path.join(
    orbit_library_path, f"turbines/{turbine_model}.yaml"
)
filename_orbit_config = os.path.join(
    orbit_library_path, f"plant/orbit-config-{turbine_model}-stripped.yaml"
)
filename_floris_config = os.path.join(
    orbit_library_path, f"floris/floris_input_{turbine_model}.yaml"
)
filename_greenheart_config = os.path.join(
    orbit_library_path, f"plant/greenheart_config.yaml"
)
filename_greenheart_config_onshore = os.path.join(
        orbit_library_path, f"plant/greenheart_config_onshore.yaml"
)
filename_hopp_config = os.path.join(
    orbit_library_path, f"plant/hopp_config.yaml"
)
filename_hopp_config_wind_wave = os.path.join(
        orbit_library_path, f"plant/hopp_config_wind_wave.yaml"
)
filename_hopp_config_wind_wave_solar = os.path.join(
        orbit_library_path, f"plant/hopp_config_wind_wave_solar.yaml"
)
filename_hopp_config_wind_wave_solar_battery = os.path.join(
        orbit_library_path, f"plant/hopp_config_wind_wave_solar_battery.yaml"
)

rtol = 1E-5

def test_simulation_wind(subtests):
    config = GreenHeartSimulationConfig(
        filename_hopp_config=filename_hopp_config,
        filename_greenheart_config=filename_greenheart_config,
        filename_turbine_config=filename_turbine_config,
        filename_orbit_config=filename_orbit_config,
        filename_floris_config=filename_floris_config,
        verbose=False,
        show_plots=False,
        save_plots=False,
        use_profast=True,
        post_processing=True,
        incentive_option=1,
        plant_design_scenario=1,
        output_level=5,
    )
    lcoe, lcoh, _, hi = run_simulation(config)

    with subtests.test("lcoh"):
        assert lcoh == approx(
            7.509261597149882
        )  # TODO base this test value on something

    with subtests.test("lcoe"):
        assert lcoe == approx(
            0.11273307417999466
        )  # TODO base this test value on something

    with subtests.test("energy sources"):
        expected_annual_energy_hybrid = hi.system.annual_energies.wind
        assert hi.system.annual_energies.hybrid == approx(expected_annual_energy_hybrid)

    with subtests.test("num_turbines conflict raise warning"):
        config.orbit_config["plant"]["num_turbines"] = 400
        with warns(UserWarning, match=f"The 'num_turbines' value"):
            lcoe, lcoh, _, hi = run_simulation(config)
            
    with subtests.test("depth conflict raise warning"):
        config.orbit_config["site"]["depth"] = 4000
        with warns(UserWarning, match=f"The site depth value"):
            lcoe, lcoh, _, hi = run_simulation(config)

    with subtests.test("turbine_spacing conflict raise warning"):
        config.orbit_config["plant"]["turbine_spacing"] = 400
        with warns(UserWarning, match=f"The 'turbine_spacing' value"):
            lcoe, lcoh, _, hi = run_simulation(config)

    with subtests.test("row_spacing conflict raise warning"):
        config.orbit_config["plant"]["row_spacing"] = 400
        with warns(UserWarning, match=f"The 'row_spacing' value"):
            lcoe, lcoh, _, hi = run_simulation(config)

def test_simulation_wind_wave(subtests):

    config = GreenHeartSimulationConfig(
        filename_hopp_config=filename_hopp_config_wind_wave,
        filename_greenheart_config=filename_greenheart_config,
        filename_turbine_config=filename_turbine_config,
        filename_orbit_config=filename_orbit_config,
        filename_floris_config=filename_floris_config,
        verbose=False,
        show_plots=False,
        save_plots=False,
        use_profast=True,
        post_processing=True,
        incentive_option=1,
        plant_design_scenario=1,
        output_level=5,
    )

    lcoe, lcoh, _, hi = run_simulation(config)

    # TODO base this test value on something
    with subtests.test("lcoh"):
        assert lcoh == approx(8.450558037665157, rel=rtol)

    # prior to 20240207 value was approx(0.11051228251811765) # TODO base this test value on something
    with subtests.test("lcoe"):
        assert lcoe == approx(0.1327684184657075, rel=rtol)


def test_simulation_wind_wave_solar(subtests):

    config = GreenHeartSimulationConfig(
        filename_hopp_config=filename_hopp_config_wind_wave_solar,
        filename_greenheart_config=filename_greenheart_config,
        filename_turbine_config=filename_turbine_config,
        filename_orbit_config=filename_orbit_config,
        filename_floris_config=filename_floris_config,
        verbose=False,
        show_plots=False,
        save_plots=False,
        use_profast=True,
        post_processing=True,
        incentive_option=1,
        plant_design_scenario=11,
        output_level=5,
    )

    lcoe, lcoh, _, hi = run_simulation(config)

    # prior to 20240207 value was approx(10.823798551850347)
    # TODO base this test value on something. Currently just based on output at writing.
    with subtests.test("lcoh"):
        assert lcoh == approx(12.946208023181846, rel=rtol)

    # prior to 20240207 value was approx(0.11035426429749774)
    # TODO base this test value on something. Currently just based on output at writing.
    with subtests.test("lcoe"):
        assert lcoe == approx(0.1325651191367888, rel=rtol)


def test_simulation_wind_wave_solar_battery(subtests):

    config = GreenHeartSimulationConfig(
        filename_hopp_config=filename_hopp_config_wind_wave_solar_battery,
        filename_greenheart_config=filename_greenheart_config,
        filename_turbine_config=filename_turbine_config,
        filename_orbit_config=filename_orbit_config,
        filename_floris_config=filename_floris_config,
        verbose=False,
        show_plots=False,
        save_plots=False,
        use_profast=True,
        post_processing=True,
        incentive_option=1,
        plant_design_scenario=10,
        output_level=8,
    )

    results = run_simulation(config)

    with subtests.test("lcoh"):
        # TODO base this test value on something. Currently just based on output at writing.
        assert results.lcoh == approx(17.315152646007785, rel=rtol)

    # TODO base this test value on something. Currently just based on output at writing.
    with subtests.test("lcoe"):
        # TODO base this test value on something. Currently just based on output at writing.
        assert results.lcoe == approx(0.13275466291379373, rel=rtol)  

    with subtests.test("no conflict in om cost does not raise warning"):
        with warnings.catch_warnings():
            warnings.simplefilter("error")

    with subtests.test("wind_om_per_kw conflict raise warning"):
        config.hopp_config["technologies"]["wind"]["fin_model"]["system_costs"]["om_capacity"][0] = 1.0
        with warns(UserWarning, match=f"The 'om_capacity' value in the wind 'fin_model'"):
            _ = run_simulation(config)
    
    with subtests.test("pv_om_per_kw conflict raise warning"):
        config.hopp_config["technologies"]["pv"]["fin_model"]["system_costs"]["om_capacity"][0] = 1.0
        with warns(UserWarning, match=f"The 'om_capacity' value in the pv 'fin_model'"):
            _ = run_simulation(config)

    with subtests.test("battery_om_per_kw conflict raise warning"):
        config.hopp_config["technologies"]["battery"]["fin_model"]["system_costs"]["om_capacity"][0] = 1.0
        with warns(UserWarning, match=f"The 'om_capacity' value in the battery 'fin_model'"):
            _ = run_simulation(config)

def test_simulation_wind_onshore(subtests):

    config = GreenHeartSimulationConfig(
        filename_hopp_config=filename_hopp_config,
        filename_greenheart_config=filename_greenheart_config_onshore,
        filename_turbine_config=filename_turbine_config,
        filename_floris_config=filename_floris_config,
        verbose=False,
        show_plots=False,
        save_plots=False,
        use_profast=True,
        post_processing=False,
        incentive_option=1,
        plant_design_scenario=9,
        output_level=5,
    )
    # based on 2023 ATB moderate case for onshore wind
    config.hopp_config["config"]["cost_info"]["wind_installed_cost_mw"] = 1434000.0 
    # based on 2023 ATB moderate case for onshore wind
    config.hopp_config["config"]["cost_info"]["wind_om_per_kw"] = 29.567
    # set skip_financial to false for onshore wind
    config.hopp_config["config"]["simulation_options"]["wind"]["skip_financial"] = False

    lcoe, lcoh, _, _ = run_simulation(config)

    # TODO base this test value on something
    with subtests.test("lcoh"):
        assert lcoh == approx(3.1691092704830357, rel=rtol)  

    # TODO base this test value on something
    with subtests.test("lcoe"):
        assert lcoe == approx(0.03486192934806013, rel=rtol)


def test_simulation_wind_onshore_steel_ammonia(subtests):

    config = GreenHeartSimulationConfig(
        filename_hopp_config=filename_hopp_config,
        filename_greenheart_config=filename_greenheart_config_onshore,
        filename_turbine_config=filename_turbine_config,
        filename_floris_config=filename_floris_config,
        verbose=False,
        show_plots=False,
        save_plots=True,
        output_dir=os.path.abspath(pathlib.Path(__file__).parent.resolve()) + "/output/",
        use_profast=True,
        post_processing=True,
        incentive_option=1,
        plant_design_scenario=9,
        output_level=7,
    )
    
    # based on 2023 ATB moderate case for onshore wind
    config.hopp_config["config"]["cost_info"]["wind_installed_cost_mw"] = 1434000.0 
    # based on 2023 ATB moderate case for onshore wind
    config.hopp_config["config"]["cost_info"]["wind_om_per_kw"] = 29.567
    config.hopp_config["technologies"]["wind"]["fin_model"]["system_costs"]["om_fixed"][0] = config.hopp_config["config"]["cost_info"]["wind_om_per_kw"]
    # set skip_financial to false for onshore wind
    config.hopp_config["config"]["simulation_options"]["wind"]["skip_financial"] = False
    lcoe, lcoh, steel_finance, ammonia_finance = run_simulation(config)

    # TODO base this test value on something
    with subtests.test("lcoh"):
        assert lcoh == approx(3.1691092704830357, rel=rtol)

    # TODO base this test value on something
    with subtests.test("lcoe"):
        assert lcoe == approx(0.03486192934806013, rel=rtol)

    # TODO base this test value on something
    with subtests.test("steel_finance"):
        lcos_expected = 1357.046163641118

        assert steel_finance.sol.get("price") == approx(lcos_expected, rel=rtol)

    # TODO base this test value on something
    with subtests.test("ammonia_finance"):
        lcoa_expected = 1.0419096226034346

        assert ammonia_finance.sol.get("price") == approx(lcoa_expected, rel=rtol)

def test_simulation_wind_battery_pv_onshore_steel_ammonia(subtests):

    plant_design_scenario = 12

    config = GreenHeartSimulationConfig(
        filename_hopp_config=filename_hopp_config_wind_wave_solar_battery,
        filename_greenheart_config=filename_greenheart_config_onshore,
        filename_turbine_config=filename_turbine_config,
        filename_orbit_config=filename_orbit_config,
        filename_floris_config=filename_floris_config,
        verbose=False,
        show_plots=False,
        save_plots=True,
        output_dir=os.path.abspath(pathlib.Path(__file__).parent.resolve()) + "/output/",
        use_profast=True,
        post_processing=True,
        incentive_option=1,
        plant_design_scenario=plant_design_scenario,
        output_level=8,
    )
    
    # based on 2023 ATB moderate case for onshore wind
    config.hopp_config["config"]["cost_info"]["wind_installed_cost_mw"] = 1434000.0 
    # based on 2023 ATB moderate case for onshore wind
    config.hopp_config["config"]["cost_info"]["wind_om_per_kw"] = 29.567
    config.hopp_config["technologies"]["wind"]["fin_model"]["system_costs"]["om_fixed"][0] = config.hopp_config["config"]["cost_info"]["wind_om_per_kw"]
    # set skip_financial to false for onshore wind
    config.hopp_config["config"]["simulation_options"]["wind"]["skip_financial"] = False
    # exclude wave
    config.hopp_config["technologies"].pop("wave")
    config.hopp_config["site"]["wave"] = False
    # colocated end-use
    config.greenheart_config["plant_design"][f"scenario{plant_design_scenario}"]["transportation"] = "colocated" 

    # run the simulation
    greenheart_output = run_simulation(config)

    # TODO base this test value on something
    with subtests.test("lcoh"):
        assert greenheart_output.lcoh == approx(3.1509756450008752, rel=rtol)

    # TODO base this test value on something
    with subtests.test("lcoe"):
        assert greenheart_output.lcoe == approx(0.03476011434910082, rel=rtol)

    # TODO base this test value on something
    with subtests.test("steel_finance"):
        lcos_expected = 1349.3364242679354

        assert greenheart_output.steel_finance.sol.get("price") == approx(lcos_expected, rel=rtol)

    # TODO base this test value on something
    with subtests.test("ammonia_finance"):
        lcoa_expected = 1.0404837286866984

        assert greenheart_output.ammonia_finance.sol.get("price") == approx(lcoa_expected, rel=rtol)

    with subtests.test("check time series lengths"):
        expected_length = 8760
        
        for key in greenheart_output.hourly_energy_breakdown.keys():
            assert len(greenheart_output.hourly_energy_breakdown[key]) == expected_length

def test_simulation_wind_onshore_steel_ammonia_ss_h2storage(subtests):

    config = GreenHeartSimulationConfig(
        filename_hopp_config=filename_hopp_config,
        filename_greenheart_config=filename_greenheart_config_onshore,
        filename_turbine_config=filename_turbine_config,
        filename_floris_config=filename_floris_config,
        verbose=False,
        show_plots=False,
        save_plots=True,
        output_dir=os.path.abspath(pathlib.Path(__file__).parent.resolve()) + "/output/",
        use_profast=True,
        post_processing=True,
        incentive_option=1,
        plant_design_scenario=9,
        output_level=7,
    )

    config.greenheart_config['h2_storage']['size_capacity_from_demand']['flag'] = True
    config.greenheart_config['h2_storage']['type'] = 'pipe'
    
    # based on 2023 ATB moderate case for onshore wind
    config.hopp_config["config"]["cost_info"]["wind_installed_cost_mw"] = 1434000.0 
    # based on 2023 ATB moderate case for onshore wind
    config.hopp_config["config"]["cost_info"]["wind_om_per_kw"] = 29.567
    config.hopp_config["technologies"]["wind"]["fin_model"]["system_costs"]["om_fixed"][0] = config.hopp_config["config"]["cost_info"]["wind_om_per_kw"]
    # set skip_financial to false for onshore wind
    config.hopp_config["config"]["simulation_options"]["wind"]["skip_financial"] = False
    lcoe, lcoh, steel_finance, ammonia_finance = run_simulation(config)

    # TODO base this test value on something
    with subtests.test("lcoh"):
        assert lcoh == approx(10.0064010897151, rel=rtol)

    # TODO base this test value on something
    with subtests.test("lcoe"):
        assert lcoe == approx(0.03486192934806013, rel=rtol)

    # TODO base this test value on something
    with subtests.test("steel_finance"):
        lcos_expected = 1812.985744428756 

        assert steel_finance.sol.get("price") == approx(lcos_expected, rel=rtol)

    # TODO base this test value on something
    with subtests.test("ammonia_finance"):
        lcoa_expected = 1.0419096226034346

        assert ammonia_finance.sol.get("price") == approx(lcoa_expected, rel=rtol)


import matplotlib.pyplot as plt


def plot_power_to_load(titletext, df_mean, df_ci, y, ylim, colors, xticks_major, xticks_minor,
                       xlabels_major, xlabels_minor,  save_location):

    fig, ax = plt.subplots(figsize=(15, 10))

    for i in y:
        if i % 24 == 0:
            ax.plot(y[i: i + 24], df_mean.pv_power_production[i: i + 24], marker="o", label="$PV Power$",
                    c=colors[1])
            ax.fill_between(
                y[i: i + 24],
                (df_mean.pv_power_production - df_ci.pv_power_production)[i: i + 24],
                (df_mean.pv_power_production + df_ci.pv_power_production)[i: i + 24],
                alpha=0.3, color=colors[1], label="$PV Power$ 95% CI"
            )

            ax.plot(y[i: i + 24], df_mean.wind_power_production[i: i + 24], marker="o",
                    label="$Wind Power$",
                    c=colors[0])
            ax.fill_between(
                y[i: i + 24],
                (df_mean.wind_power_production - df_ci.wind_power_production)[i: i + 24],
                (df_mean.wind_power_production + df_ci.wind_power_production)[i: i + 24],
                alpha=0.3, color=colors[0], label="$Wind Power$ 95% CI"
            )
            ax.plot(y[i: i + 24], df_mean.combined_pv_wind_power_production[i: i + 24], marker="o",
                    label="$Wind + PV Combined$", c=colors[3])
            ax.fill_between(
                y[i: i + 24],
                (df_mean.combined_pv_wind_power_production - df_ci.combined_pv_wind_power_production)[i: i + 24],
                (df_mean.combined_pv_wind_power_production + df_ci.combined_pv_wind_power_production)[i: i + 24],
                alpha=0.3, color=colors[3], label="$Wind + PV Combined$ 95% CI"
            )

            ax.plot(y[i: i + 24], df_mean.combined_pv_wind_storage_power_production[i: i + 24],
                    marker="o", label="$Wind + PV + Storage Combined$", c=colors[2])
            ax.fill_between(
                y[i: i + 24],
                (df_mean.combined_pv_wind_storage_power_production - df_ci.combined_pv_wind_storage_power_production)[
                i: i + 24],
                (df_mean.combined_pv_wind_storage_power_production + df_ci.combined_pv_wind_storage_power_production)[
                i: i + 24],
                alpha=0.3, color=colors[2], label="$Wind + PV + Storage Combined$ 95% CI"
            )

            ax.plot(y[i: i + 24], df_mean.storage_power_to_load[i: i + 24],
                    marker="o", label="$Storage Power$", c=colors[4])
            ax.fill_between(
                y[i: i + 24],
                (df_mean.storage_power_to_load - df_ci.storage_power_to_load)[i: i + 24],
                (df_mean.storage_power_to_load + df_ci.storage_power_to_load)[i: i + 24],
                alpha=0.3, color=colors[4], label="$Storage Power$ 95% CI"
            )

    ax.set_ylabel("Power (kW)")
    ax.set_xlim(0, 24)
    ax.set_ylim(ylim[0], ylim[1])
    ax.set_xticks(xticks_major)
    for t in ax.get_xticklabels():
        t.set_y(-0.05)
    ax.set_xticks(xticks_minor, minor=True)
    ax.set_xticklabels(xlabels_major, ha="right")
    ax.set_xticklabels(xlabels_minor, minor=True)
    ax.set_xlabel("Hour of day")

    ax.set_title(titletext)
    plt.grid(alpha=0.7)
    plt.grid(alpha=0.2, which="minor")

    handles, labels = ax.get_legend_handles_labels()
    labels_set = ["$PV Power$", "$PV Power$ 95% CI",
                  "$Wind Power$", "$Wind Power$ 95% CI",
                  "$Wind + PV Combined$", "$Wind + PV Combined$ 95% CI",
                  "$Wind + PV + Storage Combined$", "$Wind + PV + Storage Combined$ 95% CI",
                  "$Storage Power$", "$Storage Power$ 95% CI"]

    ax.grid(alpha=0.7)
    ax.grid(alpha=0.2, which="minor")

    ix_filter = [labels.index(el) for el in labels_set]
    handles = [handles[ix] for ix in ix_filter]
    labels = [labels[ix] for ix in ix_filter]
    ax.legend(handles, labels, ncol=5, loc="lower left")

    plt.tight_layout()

    # plt.show()
    plt.savefig(save_location, dpi=240, bbox_to_inches="tight")

    plt.close('all')


import pytest
from pytest import fixture

from greenheart.simulation.technologies.hydrogen.h2_storage.pipe_storage.underground_pipe_storage import UndergroundPipeStorage

# Test values are based on conclusions of Papadias 2021 and are in 2019 USD

in_dict = {
    "h2_storage_kg": 1000000,
    "system_flow_rate":100000,
    "model":'papadias',
    "compressor_output_pressure": 100}

@fixture
def pipe_storage():
    pipe_storage = UndergroundPipeStorage(in_dict)

    return pipe_storage


def test_init():
    pipe_storage = UndergroundPipeStorage(in_dict)

    assert pipe_storage.input_dict is not None
    assert pipe_storage.output_dict is not None

def test_capex_per_kg(pipe_storage):
    pipe_storage_capex_per_kg, _installed_capex, _compressor_capex = pipe_storage.pipe_storage_capex()
    assert pipe_storage_capex_per_kg == pytest.approx(512.689247292)

def test_capex(pipe_storage):
    _pipe_storage_capex_per_kg, installed_capex, _compressor_capex = pipe_storage.pipe_storage_capex()
    assert installed_capex == pytest.approx(508745483.851)

def test_compressor_capex(pipe_storage):
    _pipe_storage_capex_per_kg, _installed_capex, compressor_capex = pipe_storage.pipe_storage_capex()
    assert compressor_capex == pytest.approx(5907549.297)

def test_capex_output_dict(pipe_storage):
    _pipe_storage_capex_per_kg, _installed_capex, _compressor_capex = pipe_storage.pipe_storage_capex()
    assert pipe_storage.output_dict["pipe_storage_capex"] == pytest.approx(508745483.851)


def test_opex(pipe_storage):
    _pipe_storage_capex_per_kg, _installed_capex, _compressor_capex = pipe_storage.pipe_storage_capex()
    pipe_storage.pipe_storage_opex()
    assert pipe_storage.output_dict["pipe_storage_opex"] == pytest.approx(16439748)




from greenheart.simulation.technologies.hydrogen.h2_transport.h2_export_pipe import run_pipe_analysis
from numpy import isnan, flip, nansum

"""
Args:
    sections_distance (array[array]): array of arrays where each element of each sub-array holds the horizontal distance in m of a pipe section
    depth (float): depth of the site in m
    p_inlet (float): pipe inlet pressure in bar
    p_outlet (float): pipe outlet pressure in bar
    mass_flow_rate_inlet (float): flow rate at each inlet to the system
Returns:
    capex (float): total capital costs (USD) including labor, materials, and misc
    opex (float): annual operating costs (USD)
"""
def run_pipe_array(sections_distance, depth, p_inlet, p_outlet, mass_flow_rate):
    
    capex = 0
    opex = 0

    # loop over each string
    for i, pipe_string in enumerate(sections_distance):
        
        # initialize values for each string
        m_dot = 0
        p_drop = (p_inlet - p_outlet)/len(pipe_string)
        flow_rates = flip(mass_flow_rate[i])

        # loop over each section
        for j, section_length in enumerate(flip(pipe_string)):

            # nan represents an empty section (no pipe there, but array cannot be ragged)
            if isnan(section_length): continue

            # get mass flow rate for current section
            m_dot += flow_rates[j]

            # get outlet pressure for current section
            p_outlet_section = p_inlet - (j + 1)*p_drop

            # get number of risers for current section
            if j == len(pipe_string)-1:
                risers = 2
            else:
                risers = 1
                
            # get specs and costs for each section
            section_outputs = run_pipe_analysis(section_length, m_dot, p_inlet, p_outlet_section, depth, risers=risers)

            capex += section_outputs["total capital cost [$]"][0]
            opex += section_outputs["annual operating cost [$]"][0]

    return capex, opex

#   Assuming one pipe diameter for the pipeline
def run_pipe_array_const_diam(sections_distance, depth, p_inlet, p_outlet, mass_flow_rate):
    
    capex = 0
    opex = 0

    # loop over each string
    for i, pipe_string in enumerate(sections_distance):
        
        # Calculate maximum flow rate per pipe segment (pipe is sized to largest segment)
        m_dot = max(mass_flow_rate[i])

        #   Add up the length of the segment
        tot_length = nansum(pipe_string)

        #   Assume each full run has 2 risers
        risers = 2
        risers = len(pipe_string)+1 #Wasnt sure on this - is it 1 per turbine + 1 for the storage? - Jamie

        #   get specs and costs for each section
        section_outputs = run_pipe_analysis(tot_length, m_dot, p_inlet, p_outlet, depth, risers=risers)

        capex += section_outputs["total capital cost [$]"][0]
        opex += section_outputs["annual operating cost [$]"][0]

    return capex, opex


if __name__ == "__main__":
    sections_distance = [[2.85105454, 2.016     , 2.016     , 2.016     , 2.016     , 2.016     , 2.016     , 2.016     ],
                        [2.016     , 2.016     , 2.016     , 2.016     , 2.016     ,2.016     , 2.016     , 2.016     ],
                        [2.85105454, 2.016     , 2.016     , 2.016     ,        float("nan"), float("nan"),        float("nan"),        float("nan")]]

    L = 8                   # Length [km]
    m_dot = 1.5            # Mass flow rate [kg/s] assuming 300 MW -> 1.5 kg/s
    p_inlet = 30            # Inlet pressure [bar]
    p_outlet = 10           # Outlet pressure [bar]
    depth = 80              # depth of pipe [m]

    # capex, opex = run_pipe_array([[L, L], [L, L]], depth, p_inlet, p_outlet, [[m_dot, m_dot], [m_dot, m_dot]])

    # print("CAPEX (USD): ", capex)
    # print("OPEX (USD): ", opex)

    capex, opex = run_pipe_array_const_diam([[L, L], [L, L]], depth, p_inlet, p_outlet, [[m_dot, m_dot], [m_dot, m_dot]])

    print("CAPEX (USD): ", capex)
    print("OPEX (USD): ", opex)



from greenheart.simulation.technologies.hydrogen.h2_transport.h2_pipe_array import run_pipe_array, run_pipe_array_const_diam
from pytest import approx

# test that we the results we got when the code was recieved
class TestPipeArraySingleSection():
    L = 8                   # Length [km]
    m_dot = 1.5            # Mass flow rate [kg/s] assuming 300 MW -> 1.5 kg/s
    p_inlet = 30            # Inlet pressure [bar]
    p_outlet = 10           # Outlet pressure [bar]
    depth = 80              # depth of pipe [m]
    capex, opex = run_pipe_array([[L]], depth, p_inlet, p_outlet, [[m_dot]])
        
    def test_capex(self):
        assert self.capex == 2226256.16387454
    
    def test_opex(self):
        assert self.opex == 26047.197117332118

# TODO check the values in these test, they are gut checked for being slightly above what would be expected for a single distance, but not well determined

class TestPipeArrayMultiSection():
    L = 8                   # Length [km]
    m_dot = 1.5            # Mass flow rate [kg/s] assuming 300 MW -> 1.5 kg/s
    p_inlet = 30            # Inlet pressure [bar]
    p_outlet = 10           # Outlet pressure [bar]
    depth = 80              # depth of pipe [m]


    def test_capex(self):
        assert self.capex == 5129544.170342567
    
    def test_opex(self):
        assert self.opex == 60015.666793008044
    
    capex, opex = run_pipe_array([[L, L]], depth, p_inlet, p_outlet, [[m_dot, m_dot]])
        

class TestPipeArrayMultiSectionConstDiameter():
    L = 8                   # Length [km]
    m_dot = 1.5            # Mass flow rate [kg/s] assuming 300 MW -> 1.5 kg/s
    p_inlet = 30            # Inlet pressure [bar]
    p_outlet = 10           # Outlet pressure [bar]
    depth = 80              # depth of pipe [m]

    capex, opex = run_pipe_array_const_diam([[L, L], [L, L]], depth, p_inlet, p_outlet, [[m_dot, m_dot], [m_dot, m_dot]])
        
    def test_capex(self):
        assert self.capex == 10360272.637584394
    def test_opex(self):
        assert self.opex == 121215.18985973741
    
if __name__ == "__main__":
    test_set = TestPipeArraySingleSection()
    test_set = TestPipeArrayMultiSection()
    test_set = TestPipeArrayMultiSectionConstDiameter()
    


import os.path

import numpy as np
import numpy_financial as npf

import ProFAST  # system financial model
from ORBIT import ProjectManager
import pandas as pd

from typing import Dict, Union, Optional

from attrs import define, Factory, field

from hopp.simulation import HoppInterface


@define
class WindCostConfig:
    """
    Represents the inputs to the wind cost models

    Attributes:
        design_scenario (Dict[str, str]):
            Definition of plant subsystem locations (e.g. onshore platform, offshore, none, etc)
        hopp_config (Dict[str, float]):
            Configuration parameters for HOPP
        greenheart_config (Dict[str, float]):
            Configuration parameters for Greenheart
        orbit_config (Dict[str, float], optional):
            Required input structure for ORBIT
        turbine_config (Dict[str, float], optional):
            Configuration parameters specific to turbine
        orbit_hybrid_electrical_export_config (Dict[str, float], optional):
            Configuration parameters for hybrid electrical export in ORBIT, required if using a different substation size for the hybrid plant than for the wind plant alone
        weather (Union[list, tuple, numpy.ndarray], optional):
            Array-like of wind speeds for ORBIT to use in determining installation time and costs
    """

    design_scenario: Dict[str, str]
    hopp_config: Dict[str, float]
    greenheart_config: Dict[str, float]
    orbit_config: Optional[Dict[str, float]] = field(default={})
    turbine_config: Optional[Dict[str, float]] = field(default={})
    orbit_hybrid_electrical_export_config: Optional[Dict[str, float]] = field(
        default={}
    )
    weather: Optional[Union[list, tuple, np.ndarray]] = field(default=None)
    hopp_interface: Optional[HoppInterface] = field(default=None)


@define
class WindCostOutputs:
    """
    Represents the outputs to the wind cost models.

    Attributes:
        total_wind_cost_no_export (float):
            Total wind cost without export system costs
        total_used_export_system_costs (float):
            Total used export system costs
        annual_operating_cost_wind (float):
            Annual operating cost for wind
        installation_time (float, optional):
            Estimated installation time in months (default: 0.0)
        orbit_project (dict, optional):
            Details of the ORBIT project (default: None)
    """

    total_wind_cost_no_export: float
    annual_operating_cost_wind: float
    installation_time: float = field(default=0.0)
    total_used_export_system_costs: Optional[float] = field(default=0.0)
    orbit_project: Optional[Union[dict, ProjectManager]] = field(default=None)


def run_wind_cost_model(
    wind_cost_inputs: WindCostConfig, verbose=False
) -> WindCostOutputs:

    if wind_cost_inputs.design_scenario["wind_location"] == "offshore":

        # if per kw
        project, orbit_hybrid_electrical_export_project = run_orbit(
            wind_cost_inputs.orbit_config,
            verbose=verbose,
            weather=wind_cost_inputs.weather,
            orbit_hybrid_electrical_export_config=wind_cost_inputs.orbit_hybrid_electrical_export_config,
        )

        (
            total_wind_cost_no_export,
            total_used_export_system_costs,
        ) = breakout_export_costs_from_orbit_results(
            project,
            wind_cost_inputs.greenheart_config,
            wind_cost_inputs.design_scenario,
        )

        if orbit_hybrid_electrical_export_project is not None:
            (
                _,
                total_used_export_system_costs,
            ) = breakout_export_costs_from_orbit_results(
                orbit_hybrid_electrical_export_project,
                wind_cost_inputs.greenheart_config,
                wind_cost_inputs.design_scenario,
            )

        # WIND ONLY Total O&M expenses including fixed, variable, and capacity-based, $/year
        # use values from hybrid substation if a hybrid plant
        if orbit_hybrid_electrical_export_project is None:

            annual_operating_cost_wind = (
                max(project.monthly_opex.values()) * 12
            )  # np.average(hopp_results["hybrid_plant"].wind.om_total_expense)

        else:

            annual_operating_cost_wind = (
                max(orbit_hybrid_electrical_export_project.monthly_opex.values()) * 12
            )

        if (
            "installation_time"
            in wind_cost_inputs.greenheart_config["project_parameters"]
        ):
            installation_time = wind_cost_inputs.greenheart_config[
                "project_parameters"
            ]["installation_time"]
        else:
            installation_time = (project.installation_time / (365 * 24)) * (12.0 / 1.0)

        # if total amount
        # TODO
        return WindCostOutputs(
            total_wind_cost_no_export=total_wind_cost_no_export,
            total_used_export_system_costs=total_used_export_system_costs,
            annual_operating_cost_wind=annual_operating_cost_wind,
            installation_time=installation_time,
            orbit_project=project,
        )
    elif wind_cost_inputs.design_scenario["wind_location"] == "onshore":
        total_wind_cost_no_export = (
            wind_cost_inputs.hopp_config["config"]["cost_info"][
                "wind_installed_cost_mw"
            ]
            * wind_cost_inputs.hopp_config["technologies"]["wind"]["num_turbines"]
            * wind_cost_inputs.turbine_config["turbine_rating"]
        )

        annual_operating_cost_wind = wind_cost_inputs.hopp_interface.system.wind.om_total_expense[
            0
        ]

        if (
            "installation_time"
            in wind_cost_inputs.greenheart_config["project_parameters"]
        ):
            installation_time = wind_cost_inputs.greenheart_config[
                "project_parameters"
            ]["installation_time"]
        else:
            installation_time = 0

        return WindCostOutputs(
            total_wind_cost_no_export=total_wind_cost_no_export,
            annual_operating_cost_wind=annual_operating_cost_wind,
            installation_time=installation_time,
        )
    else:
        raise ValueError(
            "Wind design location must either be 'onshore' or 'offshore', but currently "
            f"'wind_location' is set to {wind_cost_inputs.design_scenario['wind_location']}."
        )


# Function to run orbit from provided inputs - this is just for wind costs
def run_orbit(
    orbit_config, verbose=False, weather=None, orbit_hybrid_electrical_export_config={}
):
    # set up ORBIT
    project = ProjectManager(orbit_config, weather=weather)

    # run ORBIT
    project.run(availability=orbit_config["installation_availability"])

    # run ORBIT for hybrid substation if applicable
    if orbit_hybrid_electrical_export_config == {}:
        hybrid_substation_project = None
    else:
        hybrid_substation_project = ProjectManager(
            orbit_hybrid_electrical_export_config, weather=weather
        )
        hybrid_substation_project.run(
            availability=orbit_config["installation_availability"]
        )

    # print results if desired
    if verbose:
        print(f"Installation CapEx:  {project.installation_capex/1e6:.0f} M")
        print(f"System CapEx:        {project.system_capex/1e6:.0f} M")
        print(f"Turbine CapEx:       {project.turbine_capex/1e6:.0f} M")
        print(f"Soft CapEx:          {project.soft_capex/1e6:.0f} M")
        print(f"Total CapEx:        {project.total_capex/1e6:.0f} M")
        print(f"Annual OpEx Rate:        {max(project.monthly_opex.values())*12:.0f} ")
        print(f"\nInstallation Time: {project.installation_time:.0f} h")
        print("\nN Substations: ", (project.phases["ElectricalDesign"].num_substations))
        print("N cables: ", (project.phases["ElectricalDesign"].num_cables))
        print("\n")

        # cable cost breakdown
        print("Cable specific costs")
        print(
            "Export cable installation CAPEX: %.2f M USD"
            % (project.phases["ExportCableInstallation"].installation_capex * 1e-6)
        )
        print("\n")

    return project, hybrid_substation_project


def adjust_orbit_costs(orbit_project, greenheart_config):

    if ("expected_plant_cost" in greenheart_config["finance_parameters"]["wind"]) and (
        greenheart_config["finance_parameters"]["wind"]["expected_plant_cost"] != "none"
    ):
        wind_capex_multiplier = (
            greenheart_config["finance_parameters"]["wind"]["expected_plant_cost"] * 1e9
        ) / orbit_project.total_capex
    else:
        wind_capex_multiplier = 1.0

    wind_total_capex = orbit_project.total_capex * wind_capex_multiplier
    wind_capex_breakdown = orbit_project.capex_breakdown
    for key in wind_capex_breakdown.keys():
        wind_capex_breakdown[key] *= wind_capex_multiplier

    return wind_total_capex, wind_capex_breakdown, wind_capex_multiplier


def breakout_export_costs_from_orbit_results(
    orbit_project, greenheart_config, design_scenario
):
    # adjust wind capex to meet expectations
    wind_total_capex, wind_capex_breakdown, wind_capex_multiplier = adjust_orbit_costs(
        orbit_project=orbit_project, greenheart_config=greenheart_config
    )

    # onshore substation cost is not included in ORBIT costs by default, so we have to add it separately
    total_wind_installed_costs_with_export = wind_total_capex

    # breakout export system costs
    array_cable_equipment_cost = wind_capex_breakdown["Array System"]
    array_cable_installation_cost = wind_capex_breakdown["Array System Installation"]
    total_array_cable_system_capex = (
        array_cable_equipment_cost + array_cable_installation_cost
    )

    export_cable_equipment_cost = wind_capex_breakdown[
        "Export System"
    ]  # this should include the onshore substation
    export_cable_installation_cost = wind_capex_breakdown["Export System Installation"]
    substation_equipment_cost = wind_capex_breakdown["Offshore Substation"]
    substation_installation_cost = wind_capex_breakdown[
        "Offshore Substation Installation"
    ]
    total_export_cable_system_capex = (
        export_cable_equipment_cost + export_cable_installation_cost
    )

    total_offshore_substation_capex = (
        substation_equipment_cost + substation_installation_cost
    )

    total_electrical_export_system_cost = (
        total_array_cable_system_capex
        + total_offshore_substation_capex
        + total_export_cable_system_capex
    )

    ## adjust wind cost to remove export
    if design_scenario["transportation"] == "hvdc+pipeline":
        unused_export_system_cost = 0.0
    elif (
        design_scenario["electrolyzer_location"] == "turbine"
        and design_scenario["h2_storage_location"] == "turbine"
    ):
        unused_export_system_cost = (
            total_array_cable_system_capex
            + total_export_cable_system_capex
            + total_offshore_substation_capex
        )
    elif (
        design_scenario["electrolyzer_location"] == "turbine"
        and design_scenario["h2_storage_location"] == "platform"
    ):
        unused_export_system_cost = (
            total_export_cable_system_capex  # TODO check assumptions here
        )
    elif (
        design_scenario["electrolyzer_location"] == "platform"
        and design_scenario["h2_storage_location"] == "platform"
    ):
        unused_export_system_cost = (
            total_export_cable_system_capex  # TODO check assumptions here
        )
    elif (
        design_scenario["electrolyzer_location"] == "platform"
        or design_scenario["electrolyzer_location"] == "turbine"
    ) and design_scenario["h2_storage_location"] == "onshore":
        unused_export_system_cost = (
            total_export_cable_system_capex  # TODO check assumptions here
        )
    else:
        unused_export_system_cost = 0.0

    total_used_export_system_costs = (
        total_electrical_export_system_cost - unused_export_system_cost
    )

    total_wind_cost_no_export = (
        total_wind_installed_costs_with_export - total_used_export_system_costs
    )

    return total_wind_cost_no_export, total_used_export_system_costs


def run_capex(
    hopp_results,
    wind_cost_results,
    electrolyzer_cost_results,
    h2_pipe_array_results,
    h2_transport_compressor_results,
    h2_transport_pipe_results,
    h2_storage_results,
    hopp_config,
    greenheart_config,
    design_scenario,
    desal_results,
    platform_results,
    verbose=False,
):

    # total_wind_cost_no_export, total_used_export_system_costs = breakout_export_costs_from_orbit_results(orbit_project, greenheart_config, design_scenario)

    # if orbit_hybrid_electrical_export_project is not None:
    #     _, total_used_export_system_costs = breakout_export_costs_from_orbit_results(orbit_hybrid_electrical_export_project, greenheart_config, design_scenario)

    # wave capex
    if hopp_config["site"]["wave"]:
        cost_dict = hopp_results["hybrid_plant"].wave.mhk_costs.cost_outputs

        wcapex = (
            cost_dict["structural_assembly_cost_modeled"]
            + cost_dict["power_takeoff_system_cost_modeled"]
            + cost_dict["mooring_found_substruc_cost_modeled"]
        )
        wbos = (
            cost_dict["development_cost_modeled"]
            + cost_dict["eng_and_mgmt_cost_modeled"]
            + cost_dict["plant_commissioning_cost_modeled"]
            + cost_dict["site_access_port_staging_cost_modeled"]
            + cost_dict["assembly_and_install_cost_modeled"]
            + cost_dict["other_infrastructure_cost_modeled"]
        )
        welec_infrastruc_costs = (
            cost_dict["array_cable_system_cost_modeled"]
            + cost_dict["export_cable_system_cost_modeled"]
            + cost_dict["other_elec_infra_cost_modeled"]
        )  # +\
        # cost_dict['onshore_substation_cost_modeled']+\
        # cost_dict['offshore_substation_cost_modeled']
        # financial = cost_dict['project_contingency']+\
        # cost_dict['insurance_during_construction']+\
        # cost_dict['reserve_accounts']
        wave_capex = wcapex + wbos + welec_infrastruc_costs
    else:
        wave_capex = 0.0

    # solar capex
    if "pv" in hopp_config["technologies"].keys():
        solar_capex = hopp_results["hybrid_plant"].pv.total_installed_cost
    else:
        solar_capex = 0.0

    # battery capex
    if "battery" in hopp_config["technologies"].keys():
        battery_capex = hopp_results["hybrid_plant"].battery.total_installed_cost
    else:
        battery_capex = 0.0

    # TODO bos capex
    # bos_capex = hopp_results["hybrid_plant"].bos.total_installed_cost

    ## desal capex
    if desal_results != None:
        desal_capex = desal_results["desal_capex_usd"]
    else:
        desal_capex = 0.0

    ## electrolyzer capex
    electrolyzer_total_capital_cost = electrolyzer_cost_results[
        "electrolyzer_total_capital_cost"
    ]

    if (
        design_scenario["electrolyzer_location"] == "platform"
        or design_scenario["h2_storage_location"] == "platform"
        or hopp_config["site"]["solar"]
    ):
        platform_costs = platform_results["capex"]
    else:
        platform_costs = 0.0

    # h2 transport
    h2_transport_compressor_capex = h2_transport_compressor_results["compressor_capex"]
    h2_transport_pipe_capex = h2_transport_pipe_results["total capital cost [$]"][0]

    ## h2 storage
    if greenheart_config["h2_storage"]["type"] == "none":
        h2_storage_capex = 0.0
    elif (
        greenheart_config["h2_storage"]["type"] == "pipe"
    ):  # ug pipe storage model includes compression
        h2_storage_capex = h2_storage_results["storage_capex"]
    elif (
        greenheart_config["h2_storage"]["type"] == "turbine"
    ):  # ug pipe storage model includes compression
        h2_storage_capex = h2_storage_results["storage_capex"]
    elif (
        greenheart_config["h2_storage"]["type"] == "pressure_vessel"
    ):  # pressure vessel storage model includes compression
        h2_storage_capex = h2_storage_results["storage_capex"]
    elif (
        greenheart_config["h2_storage"]["type"] == "salt_cavern"
    ):  # salt cavern storage model includes compression
        h2_storage_capex = h2_storage_results["storage_capex"]
    elif (
        greenheart_config["h2_storage"]["type"] == "lined_rock_cavern"
    ):  # lined rock cavern storage model includes compression
        h2_storage_capex = h2_storage_results["storage_capex"]
    else:
        raise NotImplementedError(
            "the storage type you have indicated (%s) has not been implemented."
            % greenheart_config["h2_storage"]["type"]
        )

    # store capex component breakdown
    capex_breakdown = {
        "wind": wind_cost_results.total_wind_cost_no_export,
        "wave": wave_capex,
        "solar": solar_capex,
        "battery": battery_capex,
        "platform": platform_costs,
        "electrical_export_system": wind_cost_results.total_used_export_system_costs,
        "desal": desal_capex,
        "electrolyzer": electrolyzer_total_capital_cost,
        "h2_pipe_array": h2_pipe_array_results["capex"],
        "h2_transport_compressor": h2_transport_compressor_capex,
        "h2_transport_pipeline": h2_transport_pipe_capex,
        "h2_storage": h2_storage_capex,
    }

    # discount capex to appropriate year for unified costing
    for key in capex_breakdown.keys():
        if key == "h2_storage":
            # if design_scenario["h2_storage_location"] == "turbine" and greenheart_config["h2_storage"]["type"] == "turbine":
            #     cost_year = greenheart_config["finance_parameters"]["discount_years"][key][
            #         design_scenario["h2_storage_location"]
            #     ]
            # else:
            cost_year = greenheart_config["finance_parameters"]["discount_years"][key][
                greenheart_config["h2_storage"]["type"]
            ]
        else:
            cost_year = greenheart_config["finance_parameters"]["discount_years"][key]

        periods = greenheart_config["project_parameters"]["cost_year"] - cost_year

        capex_breakdown[key] = -npf.fv(
            greenheart_config["finance_parameters"]["costing_general_inflation"],
            periods,
            0.0,
            capex_breakdown[key],
        )

    total_system_installed_cost = sum(
        capex_breakdown[key] for key in capex_breakdown.keys()
    )

    if verbose:
        print("\nCAPEX Breakdown")
        for key in capex_breakdown.keys():
            print(key, "%.2f" % (capex_breakdown[key] * 1e-6), " M")

        print(
            "\nTotal system CAPEX: ",
            "$%.2f" % (total_system_installed_cost * 1e-9),
            " B",
        )

    return total_system_installed_cost, capex_breakdown


def run_opex(
    hopp_results,
    wind_cost_results,
    electrolyzer_cost_results,
    h2_pipe_array_results,
    h2_transport_compressor_results,
    h2_transport_pipe_results,
    h2_storage_results,
    hopp_config,
    greenheart_config,
    desal_results,
    platform_results,
    verbose=False,
    total_export_system_cost=0,
):
    # WIND ONLY Total O&M expenses including fixed, variable, and capacity-based, $/year
    # use values from hybrid substation if a hybrid plant
    # if orbit_hybrid_electrical_export_project is None:

    # wave opex
    if hopp_config["site"]["wave"]:
        cost_dict = hopp_results["hybrid_plant"].wave.mhk_costs.cost_outputs
        wave_opex = cost_dict["maintenance_cost"] + cost_dict["operations_cost"]
    else:
        wave_opex = 0.0

    # solar opex
    if "pv" in hopp_config["technologies"].keys():
        solar_opex = hopp_results["hybrid_plant"].pv.om_total_expense[
            0
        ]
        if solar_opex < 0.1:
            raise (RuntimeWarning(f"Solar OPEX returned as {solar_opex}"))
    else:
        solar_opex = 0.0

    # battery opex
    if "battery" in hopp_config["technologies"].keys():
        battery_opex = hopp_results["hybrid_plant"].battery.om_total_expense[
            0
        ]
        if battery_opex < 0.1:
            raise (RuntimeWarning(f"Battery OPEX returned as {battery_opex}"))
    else:
        battery_opex = 0.0

    # H2 OPEX
    platform_operating_costs = platform_results["opex"]  # TODO update this

    annual_operating_cost_h2 = electrolyzer_cost_results["electrolyzer_OM_cost_annual"]

    h2_transport_compressor_opex = h2_transport_compressor_results[
        "compressor_opex"
    ]  # annual

    h2_transport_pipeline_opex = h2_transport_pipe_results["annual operating cost [$]"][
        0
    ]  # annual

    storage_opex = h2_storage_results["storage_opex"]
    # desal OPEX
    if desal_results != None:
        desal_opex = desal_results["desal_opex_usd_per_year"]
    else:
        desal_opex = 0.0
    annual_operating_cost_desal = desal_opex

    # store opex component breakdown
    opex_breakdown_annual = {
        "wind_and_electrical": wind_cost_results.annual_operating_cost_wind,
        "platform": platform_operating_costs,
        #   "electrical_export_system": total_export_om_cost,
        "wave": wave_opex,
        "solar": solar_opex,
        "battery": battery_opex,
        "desal": annual_operating_cost_desal,
        "electrolyzer": annual_operating_cost_h2,
        "h2_pipe_array": h2_pipe_array_results["opex"],
        "h2_transport_compressor": h2_transport_compressor_opex,
        "h2_transport_pipeline": h2_transport_pipeline_opex,
        "h2_storage": storage_opex,
    }

    # discount opex to appropriate year for unified costing
    for key in opex_breakdown_annual.keys():
        if key == "h2_storage":
            cost_year = greenheart_config["finance_parameters"]["discount_years"][key][
                greenheart_config["h2_storage"]["type"]
            ]
        else:
            cost_year = greenheart_config["finance_parameters"]["discount_years"][key]

        periods = greenheart_config["project_parameters"]["cost_year"] - cost_year
        opex_breakdown_annual[key] = -npf.fv(
            greenheart_config["finance_parameters"]["costing_general_inflation"],
            periods,
            0.0,
            opex_breakdown_annual[key],
        )

    # Calculate the total annual OPEX of the installed system
    total_annual_operating_costs = sum(opex_breakdown_annual.values())

    if verbose:
        print("\nAnnual OPEX Breakdown")
        for key in opex_breakdown_annual.keys():
            print(key, "%.2f" % (opex_breakdown_annual[key] * 1e-6), " M")

        print(
            "\nTotal Annual OPEX: ",
            "$%.2f" % (total_annual_operating_costs * 1e-6),
            " M",
        )
        print(opex_breakdown_annual)
    return total_annual_operating_costs, opex_breakdown_annual


def run_profast_lcoe(
    greenheart_config,
    wind_cost_results,
    capex_breakdown,
    opex_breakdown,
    hopp_results,
    incentive_option,
    design_scenario,
    verbose=False,
    show_plots=False,
    save_plots=False,
    output_dir="./output/",
):
    gen_inflation = greenheart_config["finance_parameters"]["profast_general_inflation"]

    if (
        design_scenario["h2_storage_location"] == "onshore"
        or design_scenario["electrolyzer_location"] == "onshore"
    ):
        if 'land_cost' in greenheart_config['finance_parameters']:
            land_cost = greenheart_config['finance_parameters']['land_cost']
        else:
            land_cost = 1e6  # TODO should model this
    else:
        land_cost = 0.0

    pf = ProFAST.ProFAST()
    pf.set_params(
        "commodity",
        {
            "name": "electricity",
            "unit": "kWh",
            "initial price": 100,
            "escalation": gen_inflation,
        },
    )
    pf.set_params(
        "capacity",
        np.sum(hopp_results["combined_hybrid_power_production_hopp"]) / 365.0,
    )  # kWh/day
    pf.set_params("maintenance", {"value": 0, "escalation": gen_inflation})
    pf.set_params(
        "analysis start year", greenheart_config["project_parameters"]["atb_year"] + 1
    )
    pf.set_params(
        "operating life", greenheart_config["project_parameters"]["project_lifetime"]
    )
    pf.set_params("installation months", wind_cost_results.installation_time)
    pf.set_params(
        "installation cost",
        {
            "value": 0,
            "depr type": "Straight line",
            "depr period": 4,
            "depreciable": False,
        },
    )
    if land_cost > 0:
        pf.set_params("non depr assets", land_cost)
        pf.set_params(
            "end of proj sale non depr assets",
            land_cost
            * (1 + gen_inflation)
            ** greenheart_config["project_parameters"]["project_lifetime"],
        )
    pf.set_params("demand rampup", 0)
    pf.set_params("long term utilization", 1)
    pf.set_params("credit card fees", 0)
    pf.set_params(
        "sales tax", greenheart_config["finance_parameters"]["sales_tax_rate"]
    )
    pf.set_params("license and permit", {"value": 00, "escalation": gen_inflation})
    pf.set_params("rent", {"value": 0, "escalation": gen_inflation})
    pf.set_params(
        "property tax and insurance",
        greenheart_config["finance_parameters"]["property_tax"]
        + greenheart_config["finance_parameters"]["property_insurance"],
    )
    pf.set_params(
        "admin expense",
        greenheart_config["finance_parameters"][
            "administrative_expense_percent_of_sales"
        ],
    )
    pf.set_params(
        "total income tax rate",
        greenheart_config["finance_parameters"]["total_income_tax_rate"],
    )
    pf.set_params(
        "capital gains tax rate",
        greenheart_config["finance_parameters"]["capital_gains_tax_rate"],
    )
    pf.set_params("sell undepreciated cap", True)
    pf.set_params("tax losses monetized", True)
    pf.set_params("general inflation rate", gen_inflation)
    pf.set_params(
        "leverage after tax nominal discount rate",
        greenheart_config["finance_parameters"]["discount_rate"],
    )
    pf.set_params(
        "debt equity ratio of initial financing",
        (
            greenheart_config["finance_parameters"]["debt_equity_split"]
            / (100 - greenheart_config["finance_parameters"]["debt_equity_split"])
        ),
    )
    pf.set_params("debt type", greenheart_config["finance_parameters"]["debt_type"])
    pf.set_params(
        "loan period if used", greenheart_config["finance_parameters"]["loan_period"]
    )
    pf.set_params(
        "debt interest rate",
        greenheart_config["finance_parameters"]["debt_interest_rate"],
    )
    pf.set_params(
        "cash onhand", greenheart_config["finance_parameters"]["cash_onhand_months"]
    )

    # ----------------------------------- Add capital items to ProFAST ----------------
    if "wind" in capex_breakdown.keys():
        pf.add_capital_item(
            name="Wind System",
            cost=capex_breakdown["wind"],
            depr_type=greenheart_config["finance_parameters"]["depreciation_method"],
            depr_period=greenheart_config["finance_parameters"]["depreciation_period"],
            refurb=[0],
        )
    if "wave" in capex_breakdown.keys():
        pf.add_capital_item(
            name="Wave System",
            cost=capex_breakdown["wave"],
            depr_type=greenheart_config["finance_parameters"]["depreciation_method"],
            depr_period=greenheart_config["finance_parameters"]["depreciation_period"],
            refurb=[0],
        )

    if "solar" in capex_breakdown.keys():
        pf.add_capital_item(
            name="Solar System",
            cost=capex_breakdown["solar"],
            depr_type=greenheart_config["finance_parameters"]["depreciation_method"],
            depr_period=greenheart_config["finance_parameters"]["depreciation_period"],
            refurb=[0],
        )

    if "battery" in capex_breakdown.keys():
        pf.add_capital_item(
            name="Battery System",
            cost=capex_breakdown["battery"],
            depr_type=greenheart_config["finance_parameters"]["depreciation_method"],
            depr_period=greenheart_config["finance_parameters"]["depreciation_period"],
            refurb=[0],
        )

    if design_scenario["transportation"] == "hvdc+pipeline" or not (
        design_scenario["electrolyzer_location"] == "turbine"
        and design_scenario["h2_storage_location"] == "turbine"
    ):
        pf.add_capital_item(
            name="Electrical Export system",
            cost=capex_breakdown["electrical_export_system"],
            depr_type=greenheart_config["finance_parameters"]["depreciation_method"],
            depr_period=greenheart_config["finance_parameters"]["depreciation_period"],
            refurb=[0],
        )

    # -------------------------------------- Add fixed costs--------------------------------
    pf.add_fixed_cost(
        name="Wind and Electrical Fixed O&M Cost",
        usage=1.0,
        unit="$/year",
        cost=opex_breakdown["wind_and_electrical"],
        escalation=gen_inflation,
    )

    if "wave" in opex_breakdown.keys():
        pf.add_fixed_cost(
            name="Wave O&M Cost",
            usage=1.0,
            unit="$/year",
            cost=opex_breakdown["wave"],
            escalation=gen_inflation,
        )

    if "solar" in opex_breakdown.keys():
        pf.add_fixed_cost(
            name="Solar O&M Cost",
            usage=1.0,
            unit="$/year",
            cost=opex_breakdown["solar"],
            escalation=gen_inflation,
        )

    if "battery" in opex_breakdown.keys():
        pf.add_fixed_cost(
            name="Battery O&M Cost",
            usage=1.0,
            unit="$/year",
            cost=opex_breakdown["battery"],
            escalation=gen_inflation,
        )

    # ------------------------------------- add incentives -----------------------------------
    """ Note: ptc units must be given to ProFAST in terms of dollars per unit of the primary commodity being produced
        Note: full tech-nutral (wind) tax credits are no longer available if constructions starts after Jan. 1 2034 (Jan 1. 2033 for h2 ptc)"""

    # catch incentive option and add relevant incentives
    incentive_dict = greenheart_config["policy_parameters"][
        "option%s" % (incentive_option)
    ]
    # add electricity_ptc ($/kW)
    # adjust from 1992 dollars to start year
    wind_ptc_in_dollars_per_kw = -npf.fv(
        greenheart_config['finance_parameters']['costing_general_inflation'],
        greenheart_config["project_parameters"]["atb_year"]
        + round((wind_cost_results.installation_time / 12))
        - 1992,
        0,
        incentive_dict["electricity_ptc"],
    )  # given in 1992 dollars but adjust for inflation

    pf.add_incentive(
        name="Electricity PTC",
        value=wind_ptc_in_dollars_per_kw,
        decay=-gen_inflation,
        sunset_years=10,
        tax_credit=True,
    )  # TODO check decay

    sol = pf.solve_price()

    lcoe = sol["price"]

    if verbose:
        print("\nProFAST LCOE: ", "%.2f" % (lcoe * 1e3), "$/MWh")

    if show_plots or save_plots:
        savepath = output_dir + "figures/wind_only/"
        if not os.path.exists(savepath):
            os.makedirs(savepath)
        pf.plot_costs_yearly(
            per_kg=False,
            scale="M",
            remove_zeros=True,
            remove_depreciation=False,
            fileout=savepath
            + "annual_cash_flow_wind_only_%i.png" % (design_scenario["id"]),
            show_plot=show_plots,
        )
        pf.plot_costs_yearly2(
            per_kg=False,
            scale="M",
            remove_zeros=True,
            remove_depreciation=False,
            fileout=savepath
            + "annual_cash_flow_wind_only_%i.html" % (design_scenario["id"]),
            show_plot=show_plots,
        )
        pf.plot_capital_expenses(
            fileout=savepath + "capital_expense_only_%i.png" % (design_scenario["id"]),
            show_plot=show_plots,
        )
        pf.plot_cashflow(
            fileout=savepath + "cash_flow_wind_only_%i.png" % (design_scenario["id"]),
            show_plot=show_plots,
        )
        pf.plot_costs(
            fileout=savepath + "cost_breakdown_%i.png" % (design_scenario["id"]),
            show_plot=show_plots,
        )

    return lcoe, pf


def run_profast_grid_only(
    greenheart_config,
    wind_cost_results,
    electrolyzer_physics_results,
    capex_breakdown,
    opex_breakdown,
    hopp_results,
    design_scenario,
    total_accessory_power_renewable_kw,
    total_accessory_power_grid_kw,
    verbose=False,
    show_plots=False,
    save_plots=False,
    output_dir="./output/",
):
    gen_inflation = greenheart_config["finance_parameters"]["profast_general_inflation"]

    if (
        design_scenario["h2_storage_location"] == "onshore"
        or design_scenario["electrolyzer_location"] == "onshore"
    ):
        if 'land_cost' in greenheart_config['finance_parameters']:
            land_cost = greenheart_config['finance_parameters']['land_cost']
        else:
            land_cost = 1e6  # TODO should model this
    else:
        land_cost = 0.0

    pf = ProFAST.ProFAST()
    pf.set_params(
        "commodity",
        {
            "name": "Hydrogen",
            "unit": "kg",
            "initial price": 100,
            "escalation": gen_inflation,
        },
    )
    pf.set_params(
        "capacity",
        electrolyzer_physics_results["H2_Results"][
            "Life: Annual H2 production [kg/year]"
        ]
        / 365.0,
    )  # kg/day
    pf.set_params("maintenance", {"value": 0, "escalation": gen_inflation})
    pf.set_params(
        "analysis start year", greenheart_config["project_parameters"]["atb_year"] + 1
    )
    pf.set_params(
        "operating life", greenheart_config["project_parameters"]["project_lifetime"]
    )
    pf.set_params(
        "installation cost",
        {
            "value": 0,
            "depr type": "Straight line",
            "depr period": 4,
            "depreciable": False,
        },
    )
    if land_cost > 0:
        pf.set_params("non depr assets", land_cost)
        pf.set_params(
            "end of proj sale non depr assets",
            land_cost
            * (1 + gen_inflation)
            ** greenheart_config["project_parameters"]["project_lifetime"],
        )
    pf.set_params("demand rampup", 0)
    pf.set_params("long term utilization", 1)
    pf.set_params("credit card fees", 0)
    pf.set_params(
        "sales tax", greenheart_config["finance_parameters"]["sales_tax_rate"]
    )
    pf.set_params("license and permit", {"value": 00, "escalation": gen_inflation})
    pf.set_params("rent", {"value": 0, "escalation": gen_inflation})
    pf.set_params(
        "property tax and insurance",
        greenheart_config["finance_parameters"]["property_tax"]
        + greenheart_config["finance_parameters"]["property_insurance"],
    )
    pf.set_params(
        "admin expense",
        greenheart_config["finance_parameters"][
            "administrative_expense_percent_of_sales"
        ],
    )
    pf.set_params(
        "total income tax rate",
        greenheart_config["finance_parameters"]["total_income_tax_rate"],
    )
    pf.set_params(
        "capital gains tax rate",
        greenheart_config["finance_parameters"]["capital_gains_tax_rate"],
    )
    pf.set_params("sell undepreciated cap", True)
    pf.set_params("tax losses monetized", True)
    pf.set_params("general inflation rate", gen_inflation)
    pf.set_params(
        "leverage after tax nominal discount rate",
        greenheart_config["finance_parameters"]["discount_rate"],
    )
    pf.set_params(
        "debt equity ratio of initial financing",
        (
            greenheart_config["finance_parameters"]["debt_equity_split"]
            / (100 - greenheart_config["finance_parameters"]["debt_equity_split"])
        ),
    )
    pf.set_params("debt type", greenheart_config["finance_parameters"]["debt_type"])
    pf.set_params(
        "loan period if used", greenheart_config["finance_parameters"]["loan_period"]
    )
    pf.set_params(
        "debt interest rate",
        greenheart_config["finance_parameters"]["debt_interest_rate"],
    )
    pf.set_params(
        "cash onhand", greenheart_config["finance_parameters"]["cash_onhand_months"]
    )

    # ----------------------------------- Add capital items to ProFAST ----------------
    # pf.add_capital_item(name="Wind System",cost=capex_breakdown["wind"], depr_type=greenheart_config["finance_parameters"]["depreciation_method"], depr_period=greenheart_config["finance_parameters"]["depreciation_period"],refurb=[0])
    # pf.add_capital_item(
    #     name="Electrical Export system",
    #     cost=capex_breakdown["electrical_export_system"],
    #     depr_type=greenheart_config["finance_parameters"]["depreciation_method"],
    #     depr_period=greenheart_config["finance_parameters"]["depreciation_period"],
    #     refurb=[0],
    # )

    electrolyzer_refurbishment_schedule = np.zeros(
        greenheart_config["project_parameters"]["project_lifetime"]
    )
    # refurb_period = round(
    #     greenheart_config["electrolyzer"]["time_between_replacement"] / (24 * 365)
    # )
    refurb_period = round(electrolyzer_physics_results['H2_Results']['Time Until Replacement [hrs]']/ (24 * 365)
    )
    electrolyzer_refurbishment_schedule[
        refurb_period : greenheart_config["project_parameters"][
            "project_lifetime"
        ] : refurb_period
    ] = greenheart_config["electrolyzer"]["replacement_cost_percent"]

    pf.add_capital_item(
        name="Electrolysis System",
        cost=capex_breakdown["electrolyzer"],
        depr_type=greenheart_config["finance_parameters"]["depreciation_method"],
        depr_period=greenheart_config["finance_parameters"][
            "depreciation_period_electrolyzer"
        ],
        refurb=list(electrolyzer_refurbishment_schedule),
    )

    pf.add_capital_item(
        name="Hydrogen Storage System",
        cost=capex_breakdown["h2_storage"],
        depr_type=greenheart_config["finance_parameters"]["depreciation_method"],
        depr_period=greenheart_config["finance_parameters"][
            "depreciation_period_electrolyzer"
        ],
        refurb=[0],
    )
    # pf.add_capital_item(name ="Desalination system",cost=capex_breakdown["desal"], depr_type=greenheart_config["finance_parameters"]["depreciation_method"],depr_period=greenheart_config["finance_parameters"]["depreciation_period"],refurb=[0])

    # -------------------------------------- Add fixed costs--------------------------------
    # pf.add_fixed_cost(name="Wind Fixed O&M Cost",usage=1.0, unit='$/year',cost=opex_breakdown["wind"],escalation=gen_inflation)
    # pf.add_fixed_cost(name="Electrical Export Fixed O&M Cost", usage=1.0,unit='$/year',cost=opex_breakdown["electrical_export_system"],escalation=gen_inflation)
    # pf.add_fixed_cost(name="Desalination Fixed O&M Cost",usage=1.0, unit='$/year',cost=opex_breakdown["desal"],escalation=gen_inflation)
    pf.add_fixed_cost(
        name="Electrolyzer Fixed O&M Cost",
        usage=1.0,
        unit="$/year",
        cost=opex_breakdown["electrolyzer"],
        escalation=gen_inflation,
    )
    pf.add_fixed_cost(
        name="Hydrogen Storage Fixed O&M Cost",
        usage=1.0,
        unit="$/year",
        cost=opex_breakdown["h2_storage"],
        escalation=gen_inflation,
    )

    # ---------------------- Add feedstocks, note the various cost options-------------------
    galperkg = 3.785411784
    pf.add_feedstock(
        name="Water",
        usage=sum(
            electrolyzer_physics_results["H2_Results"][
                "Water Hourly Consumption [kg/hr]"
            ]
        )
        * galperkg
        / electrolyzer_physics_results["H2_Results"][
            "Life: Annual H2 production [kg/year]"
        ],
        unit="gal",
        cost="US Average",
        escalation=gen_inflation,
    )

    # if greenheart_config["project_parameters"]["grid_connection"]:

    energy_purchase = (
        365 * 24 * greenheart_config["electrolyzer"]["rating"] * 1e3
        + sum(total_accessory_power_renewable_kw)
        + total_accessory_power_grid_kw
    )

    pf.add_fixed_cost(
        name="Electricity from grid",
        usage=1.0,
        unit="$/year",
        cost=energy_purchase * greenheart_config["project_parameters"]["ppa_price"],
        escalation=gen_inflation,
    )

    sol = pf.solve_price()

    lcoh = sol["price"]
    if verbose:
        print("\nLCOH grid only: ", "%.2f" % (lcoh), "$/kg")
        print("ProFAST grid only NPV: ", "%.2f" % (sol["NPV"]))
        print("ProFAST grid only IRR: ", "%.5f" % (max(sol["irr"])))
        print("ProFAST grid only LCO: ", "%.2f" % (sol["lco"]), "$/kg")
        print("ProFAST grid only Profit Index: ", "%.2f" % (sol["profit index"]))
        print("ProFAST grid only payback period: ", sol["investor payback period"])

    if save_plots or show_plots:
        savepaths = [
            output_dir + "figures/capex/",
            output_dir + "figures/annual_cash_flow/",
            output_dir + "figures/lcoh_breakdown/",
            output_dir + "data/",
        ]
        for savepath in savepaths:
            if not os.path.exists(savepath):
                os.makedirs(savepath)

        pf.plot_capital_expenses(
            fileout=savepaths[0]
            + "capital_expense_grid_only_%i.pdf" % (design_scenario["id"]),
            show_plot=show_plots,
        )
        pf.plot_cashflow(
            fileout=savepaths[1]
            + "cash_flow_grid_only_%i.png" % (design_scenario["id"]),
            show_plot=show_plots,
        )

        pd.DataFrame.from_dict(data=pf.cash_flow_out, orient="index").to_csv(
            savepaths[3] + "cash_flow_grid_only_%i.csv" % (design_scenario["id"])
        )

        pf.plot_costs(
            savepaths[2] + "lcoh_grid_only_%i" % (design_scenario["id"]),
            show_plot=show_plots,
        )
    return lcoh, pf


def run_profast_full_plant_model(
    greenheart_config,
    wind_cost_results,
    electrolyzer_physics_results,
    capex_breakdown,
    opex_breakdown,
    hopp_results,
    incentive_option,
    design_scenario,
    total_accessory_power_renewable_kw,
    total_accessory_power_grid_kw,
    verbose=False,
    show_plots=False,
    save_plots=False,
    output_dir="./output/",
):
    gen_inflation = greenheart_config["finance_parameters"]["profast_general_inflation"]

    if (
        design_scenario["h2_storage_location"] == "onshore"
        or design_scenario["electrolyzer_location"] == "onshore"
    ):
        if 'land_cost' in greenheart_config['finance_parameters']:
            land_cost = greenheart_config['finance_parameters']['land_cost']
        else:
            land_cost = 1e6  # TODO should model this
    else:
        land_cost = 0.0

    pf = ProFAST.ProFAST()
    pf.set_params(
        "commodity",
        {
            "name": "Hydrogen",
            "unit": "kg",
            "initial price": 100,
            "escalation": gen_inflation,
        },
    )
    pf.set_params(
        "capacity",
        electrolyzer_physics_results["H2_Results"][
            "Life: Annual H2 production [kg/year]"
        ]
        / 365.0,
    )  # kg/day
    pf.set_params("maintenance", {"value": 0, "escalation": gen_inflation})
    pf.set_params(
        "analysis start year",
        greenheart_config["project_parameters"]["atb_year"]
        + 2,  # Add financial analysis start year
    )
    pf.set_params(
        "operating life", greenheart_config["project_parameters"]["project_lifetime"]
    )
    pf.set_params(
        "installation months",
        wind_cost_results.installation_time,  # Add installation time to yaml default=0
    )
    pf.set_params(
        "installation cost",
        {
            "value": 0,
            "depr type": "Straight line",
            "depr period": 4,
            "depreciable": False,
        },
    )
    if land_cost > 0:
        pf.set_params("non depr assets", land_cost)
        pf.set_params(
            "end of proj sale non depr assets",
            land_cost
            * (1 + gen_inflation)
            ** greenheart_config["project_parameters"]["project_lifetime"],
        )
    pf.set_params("demand rampup", 0)
    pf.set_params("long term utilization", 1)  # TODO should use utilization
    pf.set_params("credit card fees", 0)
    pf.set_params(
        "sales tax", greenheart_config["finance_parameters"]["sales_tax_rate"]
    )
    pf.set_params("license and permit", {"value": 00, "escalation": gen_inflation})
    pf.set_params("rent", {"value": 0, "escalation": gen_inflation})
    # TODO how to handle property tax and insurance for fully offshore?
    pf.set_params(
        "property tax and insurance",
        greenheart_config["finance_parameters"]["property_tax"]
        + greenheart_config["finance_parameters"]["property_insurance"],
    )
    pf.set_params(
        "admin expense",
        greenheart_config["finance_parameters"][
            "administrative_expense_percent_of_sales"
        ],
    )
    pf.set_params(
        "total income tax rate",
        greenheart_config["finance_parameters"]["total_income_tax_rate"],
    )
    pf.set_params(
        "capital gains tax rate",
        greenheart_config["finance_parameters"]["capital_gains_tax_rate"],
    )
    pf.set_params("sell undepreciated cap", True)
    pf.set_params("tax losses monetized", True)
    pf.set_params("general inflation rate", gen_inflation)
    pf.set_params(
        "leverage after tax nominal discount rate",
        greenheart_config["finance_parameters"]["discount_rate"],
    )
    if greenheart_config["finance_parameters"]["debt_equity_split"]:
        pf.set_params(
            "debt equity ratio of initial financing",
            (
                greenheart_config["finance_parameters"]["debt_equity_split"]
                / (100 - greenheart_config["finance_parameters"]["debt_equity_split"])
            ),
        )  # TODO this may not be put in right
    elif greenheart_config["finance_parameters"]["debt_equity_ratio"]:
        pf.set_params(
            "debt equity ratio of initial financing",
            (greenheart_config["finance_parameters"]["debt_equity_ratio"]),
        )  # TODO this may not be put in right
    pf.set_params("debt type", greenheart_config["finance_parameters"]["debt_type"])
    pf.set_params(
        "loan period if used", greenheart_config["finance_parameters"]["loan_period"]
    )
    pf.set_params(
        "debt interest rate",
        greenheart_config["finance_parameters"]["debt_interest_rate"],
    )
    pf.set_params(
        "cash onhand", greenheart_config["finance_parameters"]["cash_onhand_months"]
    )

    # ----------------------------------- Add capital and fixed items to ProFAST ----------------
    if "wind" in capex_breakdown.keys():
        pf.add_capital_item(
            name="Wind System",
            cost=capex_breakdown["wind"],
            depr_type=greenheart_config["finance_parameters"]["depreciation_method"],
            depr_period=greenheart_config["finance_parameters"]["depreciation_period"],
            refurb=[0],
        )
    if "wave" in capex_breakdown.keys():
        pf.add_capital_item(
            name="Wave System",
            cost=capex_breakdown["wave"],
            depr_type=greenheart_config["finance_parameters"]["depreciation_method"],
            depr_period=greenheart_config["finance_parameters"]["depreciation_period"],
            refurb=[0],
        )
    if "solar" in capex_breakdown.keys():
        pf.add_capital_item(
            name="Solar System",
            cost=capex_breakdown["solar"],
            depr_type=greenheart_config["finance_parameters"]["depreciation_method"],
            depr_period=greenheart_config["finance_parameters"]["depreciation_period"],
            refurb=[0],
        )

    if "battery" in capex_breakdown.keys():
        pf.add_capital_item(
            name="Battery System",
            cost=capex_breakdown["battery"],
            depr_type=greenheart_config["finance_parameters"]["depreciation_method"],
            depr_period=greenheart_config["finance_parameters"]["depreciation_period"],
            refurb=[0],
        )

    if "platform" in capex_breakdown.keys() and capex_breakdown["platform"] > 0:
        pf.add_capital_item(
            name="Equipment Platform",
            cost=capex_breakdown["platform"],
            depr_type=greenheart_config["finance_parameters"]["depreciation_method"],
            depr_period=greenheart_config["finance_parameters"]["depreciation_period"],
            refurb=[0],
        )
        pf.add_fixed_cost(
            name="Equipment Platform O&M Cost",
            usage=1.0,
            unit="$/year",
            cost=opex_breakdown["platform"],
            escalation=gen_inflation,
        )

    pf.add_fixed_cost(
        name="Wind and Electrical Export Fixed O&M Cost",
        usage=1.0,
        unit="$/year",
        cost=opex_breakdown["wind_and_electrical"],
        escalation=gen_inflation,
    )
    if "wave" in opex_breakdown.keys():
        pf.add_fixed_cost(
            name="Wave O&M Cost",
            usage=1.0,
            unit="$/year",
            cost=opex_breakdown["wave"],
            escalation=gen_inflation,
        )

    if "solar" in opex_breakdown.keys():
        pf.add_fixed_cost(
            name="Solar O&M Cost",
            usage=1.0,
            unit="$/year",
            cost=opex_breakdown["solar"],
            escalation=gen_inflation,
        )

    if "battery" in opex_breakdown.keys():
        pf.add_fixed_cost(
            name="Battery O&M Cost",
            usage=1.0,
            unit="$/year",
            cost=opex_breakdown["battery"],
            escalation=gen_inflation,
        )

    if design_scenario["transportation"] == "hvdc+pipeline" or not (
        design_scenario["electrolyzer_location"] == "turbine"
        and design_scenario["h2_storage_location"] == "turbine"
    ):
        pf.add_capital_item(
            name="Electrical Export system",
            cost=capex_breakdown["electrical_export_system"],
            depr_type=greenheart_config["finance_parameters"]["depreciation_method"],
            depr_period=greenheart_config["finance_parameters"]["depreciation_period"],
            refurb=[0],
        )
        # TODO assess if this makes sense (electrical export O&M included in wind O&M)

    electrolyzer_refurbishment_schedule = np.zeros(
        greenheart_config["project_parameters"]["project_lifetime"]
    )
    # refurb_period = round(
    #     greenheart_config["electrolyzer"]["time_between_replacement"] / (24 * 365)
    # )
    refurb_period = round(electrolyzer_physics_results['H2_Results']['Time Until Replacement [hrs]']/ (24 * 365)
    )
    electrolyzer_refurbishment_schedule[
        refurb_period : greenheart_config["project_parameters"][
            "project_lifetime"
        ] : refurb_period
    ] = greenheart_config["electrolyzer"]["replacement_cost_percent"]

    pf.add_capital_item(
        name="Electrolysis System",
        cost=capex_breakdown["electrolyzer"],
        depr_type=greenheart_config["finance_parameters"]["depreciation_method"],
        depr_period=greenheart_config["finance_parameters"][
            "depreciation_period_electrolyzer"
        ],
        refurb=list(electrolyzer_refurbishment_schedule),
    )
    pf.add_fixed_cost(
        name="Electrolysis System Fixed O&M Cost",
        usage=1.0,
        unit="$/year",
        cost=opex_breakdown["electrolyzer"],
        escalation=gen_inflation,
    )

    if design_scenario["electrolyzer_location"] == "turbine":
        pf.add_capital_item(
            name="H2 Pipe Array System",
            cost=capex_breakdown["h2_pipe_array"],
            depr_type=greenheart_config["finance_parameters"]["depreciation_method"],
            depr_period=greenheart_config["finance_parameters"][
                "depreciation_period_electrolyzer"
            ],
            refurb=[0],
        )
        pf.add_fixed_cost(
            name="H2 Pipe Array Fixed O&M Cost",
            usage=1.0,
            unit="$/year",
            cost=opex_breakdown["h2_pipe_array"],
            escalation=gen_inflation,
        )

    if (
        (
            design_scenario["h2_storage_location"] == "onshore"
            and design_scenario["electrolyzer_location"] != "onshore"
        )
        or (
            design_scenario["h2_storage_location"] != "onshore"
            and design_scenario["electrolyzer_location"] == "onshore"
        )
        or (design_scenario["transportation"] == "hvdc+pipeline")
    ):
        pf.add_capital_item(
            name="H2 Transport Compressor System",
            cost=capex_breakdown["h2_transport_compressor"],
            depr_type=greenheart_config["finance_parameters"]["depreciation_method"],
            depr_period=greenheart_config["finance_parameters"][
                "depreciation_period_electrolyzer"
            ],
            refurb=[0],
        )
        pf.add_capital_item(
            name="H2 Transport Pipeline System",
            cost=capex_breakdown["h2_transport_pipeline"],
            depr_type=greenheart_config["finance_parameters"]["depreciation_method"],
            depr_period=greenheart_config["finance_parameters"][
                "depreciation_period_electrolyzer"
            ],
            refurb=[0],
        )

        pf.add_fixed_cost(
            name="H2 Transport Compression Fixed O&M Cost",
            usage=1.0,
            unit="$/year",
            cost=opex_breakdown["h2_transport_compressor"],
            escalation=gen_inflation,
        )
        pf.add_fixed_cost(
            name="H2 Transport Pipeline Fixed O&M Cost",
            usage=1.0,
            unit="$/year",
            cost=opex_breakdown["h2_transport_pipeline"],
            escalation=gen_inflation,
        )

    if greenheart_config["h2_storage"]["type"] != "none":
        pf.add_capital_item(
            name="Hydrogen Storage System",
            cost=capex_breakdown["h2_storage"],
            depr_type=greenheart_config["finance_parameters"]["depreciation_method"],
            depr_period=greenheart_config["finance_parameters"][
                "depreciation_period_electrolyzer"
            ],
            refurb=[0],
        )
        pf.add_fixed_cost(
            name="Hydrogen Storage Fixed O&M Cost",
            usage=1.0,
            unit="$/year",
            cost=opex_breakdown["h2_storage"],
            escalation=gen_inflation,
        )

    # ---------------------- Add feedstocks, note the various cost options-------------------
    if design_scenario["electrolyzer_location"] == "onshore":
        galperkg = 3.785411784
        pf.add_feedstock(
            name="Water",
            usage=sum(
                electrolyzer_physics_results["H2_Results"][
                    "Water Hourly Consumption [kg/hr]"
                ]
            )
            * galperkg
            / electrolyzer_physics_results["H2_Results"][
                "Life: Annual H2 production [kg/year]"
            ],
            unit="gal",
            cost="US Average",
            escalation=gen_inflation,
        )
    else:
        pf.add_capital_item(
            name="Desal System",
            cost=capex_breakdown["desal"],
            depr_type=greenheart_config["finance_parameters"]["depreciation_method"],
            depr_period=greenheart_config["finance_parameters"][
                "depreciation_period_electrolyzer"
            ],
            refurb=[0],
        )
        pf.add_fixed_cost(
            name="Desal Fixed O&M Cost",
            usage=1.0,
            unit="$/year",
            cost=opex_breakdown["desal"],
            escalation=gen_inflation,
        )

    if (
        greenheart_config["project_parameters"]["grid_connection"]
        or total_accessory_power_grid_kw > 0
    ):

        energy_purchase = sum(total_accessory_power_grid_kw) # * 365 * 24

        if greenheart_config["project_parameters"]["grid_connection"]:
            annual_energy_shortfall = np.sum(hopp_results["energy_shortfall_hopp"])
            energy_purchase += annual_energy_shortfall

        pf.add_fixed_cost(
            name="Electricity from grid",
            usage=1.0,
            unit="$/year",
            cost=energy_purchase * greenheart_config["project_parameters"]["ppa_price"],
            escalation=gen_inflation,
        )

    # ------------------------------------- add incentives -----------------------------------
    """ Note: units must be given to ProFAST in terms of dollars per unit of the primary commodity being produced
        Note: full tech-nutral (wind) tax credits are no longer available if constructions starts after Jan. 1 2034 (Jan 1. 2033 for h2 ptc)"""

    # catch incentive option and add relevant incentives
    incentive_dict = greenheart_config["policy_parameters"][
        "option%s" % (incentive_option)
    ]

    # add wind_itc (% of wind capex)
    electricity_itc_value_percent_wind_capex = incentive_dict["electricity_itc"]
    electricity_itc_value_dollars = electricity_itc_value_percent_wind_capex * (
        capex_breakdown["wind"] + capex_breakdown["electrical_export_system"]
    )
    pf.set_params(
        "one time cap inct",
        {
            "value": electricity_itc_value_dollars,
            "depr type": greenheart_config["finance_parameters"]["depreciation_method"],
            "depr period": greenheart_config["finance_parameters"][
                "depreciation_period"
            ],
            "depreciable": True,
        },
    )

    # add h2_storage_itc (% of h2 storage capex)
    itc_value_percent_h2_store_capex = incentive_dict["h2_storage_itc"]
    electricity_itc_value_dollars_h2_store = itc_value_percent_h2_store_capex * (
        capex_breakdown["h2_storage"]
    )
    pf.set_params(
        "one time cap inct",
        {
            "value": electricity_itc_value_dollars_h2_store,
            "depr type": greenheart_config["finance_parameters"]["depreciation_method"],
            "depr period": greenheart_config["finance_parameters"][
                "depreciation_period"
            ],
            "depreciable": True,
        },
    )

    # add electricity_ptc ($/kW)
    # adjust from 1992 dollars to start year
    electricity_ptc_in_dollars_per_kw = -npf.fv(
        greenheart_config['finance_parameters']['costing_general_inflation'],
        greenheart_config["project_parameters"]["atb_year"]
        + round((wind_cost_results.installation_time / 12))
        - 1992,
        0,
        incentive_dict["electricity_ptc"],
    )  # given in 1992 dollars but adjust for inflation
    kw_per_kg_h2 = (
        sum(hopp_results["combined_hybrid_power_production_hopp"])
        / electrolyzer_physics_results["H2_Results"][
            "Life: Annual H2 production [kg/year]"
        ]
    )
    electricity_ptc_in_dollars_per_kg_h2 = (
        electricity_ptc_in_dollars_per_kw * kw_per_kg_h2
    )
    pf.add_incentive(
        name="Electricity PTC",
        value=electricity_ptc_in_dollars_per_kg_h2,
        decay=-gen_inflation,
        sunset_years=10,
        tax_credit=True,
    )  # TODO check decay

    # add h2_ptc ($/kg)
    h2_ptc_inflation_adjusted = -npf.fv(
        greenheart_config['finance_parameters']['costing_general_inflation'], # use ATB year (cost inflation 2.5%) costing_general_inflation
        greenheart_config["project_parameters"]["atb_year"]
        + round((wind_cost_results.installation_time / 12))
        - 2022,
        0,
        incentive_dict["h2_ptc"],
    )
    pf.add_incentive(
        name="H2 PTC",
        value=h2_ptc_inflation_adjusted,
        decay=-gen_inflation, #correct inflation
        sunset_years=10,
        tax_credit=True,
    )  # TODO check decay

    # ------------------------------------ solve and post-process -----------------------------

    sol = pf.solve_price()

    df = pf.cash_flow_out

    lcoh = sol["price"]

    if verbose:
        print("\nProFAST LCOH: ", "%.2f" % (lcoh), "$/kg")
        print("ProFAST NPV: ", "%.2f" % (sol["NPV"]))
        print("ProFAST IRR: ", "%.5f" % (max(sol["irr"])))
        print("ProFAST LCO: ", "%.2f" % (sol["lco"]), "$/kg")
        print("ProFAST Profit Index: ", "%.2f" % (sol["profit index"]))
        print("ProFAST payback period: ", sol["investor payback period"])

        MIRR = npf.mirr(
            df["Investor cash flow"],
            greenheart_config["finance_parameters"]["debt_interest_rate"],
            greenheart_config["finance_parameters"]["discount_rate"],
        )  # TODO probably ignore MIRR
        NPV = npf.npv(
            greenheart_config["finance_parameters"]["profast_general_inflation"],
            df["Investor cash flow"],
        )
        ROI = np.sum(df["Investor cash flow"]) / abs(
            np.sum(df["Investor cash flow"][df["Investor cash flow"] < 0])
        )  # ROI is not a good way of thinking about the value of the project

        # TODO project level IRR - capex and operating cash flow

        # note: hurdle rate typically 20% IRR before investing in it due to typically optimistic assumptions
        # note: negative retained earnings (keeping debt, paying down equity) - to get around it, do another line for retained earnings and watch dividends paid by the rpoject (net income/equity should stay positive this way)

        print("Investor NPV: ", np.round(NPV * 1e-6, 2), "M USD")
        print("Investor MIRR: ", np.round(MIRR, 5), "")
        print("Investor ROI: ", np.round(ROI, 5), "")

    if save_plots or show_plots:
        savepaths = [
            output_dir + "figures/capex/",
            output_dir + "figures/annual_cash_flow/",
            output_dir + "figures/lcoh_breakdown/",
            output_dir + "data/",
        ]
        for savepath in savepaths:
            if not os.path.exists(savepath):
                os.makedirs(savepath)

        pf.plot_capital_expenses(
            fileout=savepaths[0] + "capital_expense_%i.pdf" % (design_scenario["id"]),
            show_plot=show_plots,
        )
        pf.plot_cashflow(
            fileout=savepaths[1] + "cash_flow_%i.png" % (design_scenario["id"]),
            show_plot=show_plots,
        )

        pd.DataFrame.from_dict(data=pf.cash_flow_out).to_csv(
            savepaths[3] + "cash_flow_%i.csv" % (design_scenario["id"])
        )

        pf.plot_costs(
            savepaths[2] + "lcoh_%i" % (design_scenario["id"]), show_plot=show_plots,
        )

    return lcoh, pf



import math
import numpy as np 
from matplotlib import pyplot as plt 
from greenheart.to_organize.H2_Analysis.simple_cash_annuals import simple_cash_annuals

verbose = False
class PipelineASME: 
    """    
    # ASME B31-12 Hydrogen Pipes, PL3.7.1 Steel Pipe design
    """
    def __init__(self, input_dict, output_dict):
        self.input_dict = input_dict 
        self.output_dict = output_dict
    
        #self.pipeline_model = input_dict['pipeline_model']
        self.pipe_diam_in = input_dict['pipe_diam_in']
        self.pipe_thic_in = input_dict['pipe_thic_in']
        self.dist_to_h2_load_km = input_dict['dist_to_h2_load_km']
        self.site_depth_m = input_dict['site_depth_m']
        self.h2_flow_kg_h = input_dict['flow_rate_kg_hr']
        self.pres_in_bar = input_dict['pressure_bar']
        self.steel_cost_ton = input_dict['steel_cost_ton']
        #self.offshore_param = input_dict['offshore_bool']
        #self.plant_life = input_dict['plant_life']
        #self.useful_life = input_dict['useful_life']

        # Assumptions
        self.useful_life = 30       #[years]
        self.plant_life = 30        # [years]

        # TODO: Calculate density as a function of temperature (ambient?)
        self.rho_h2 = 0.08375 # STP H2 density, kg/m**3
        self.mu_h2 = 0.88e-5 # Pa-s at 20C

        # Undersea environment 
        self.rho_h2o = 1026 # kg/m**3
        self.g = 9.81 # m/s**2 

        # Estimates from Weight/Price charts 
        # Example: https://www.savoypipinginc.com/blog/live-stock-and-current-price.html
        #self.Smys = 52000        # Min. Yield Hoop Stress (Table IX-1B)
        self.rho_steel =  7529.0 # kg/m**3 = 470.0 lb / ft**3 estimate 
        #self.cost_per_kg = 2.80 # $ per kg 
        self.Re = 2100.0
        
        # ASME B31.12 Hydrogen Pipeline parameters 
        self.F = 0.5             # Design factor (3.7.1.1)
        self.T = 1.0             # Temperature factor: < 250degF (3.7.1-3)
        self.Smin = 52000        # X52 Carbon Steel
        self.E = 1.0             # Long. Weld Factor: Seamless (Table IX-3B)
        self.Hf = 1.0            # Table IX-5b 
        self.pipe_sch = 40       # common pipe schedule for Hoop Stress estimation

        # Conversions factors
        self.in2cm = 2.54
        self.bar2psi = 14.5
        self.bar2Pa = 100000.0
        
    def pipelineDesign(self): 
        #flow_rate_kg_per_hr, inlet_pressure_bar, outlet_pressure_bar, length_m):
        """
        Calculate pressure drop based on General Flow Equation (Darcy eq.)
        Inputs
        ======
         gas flow rate (kg/hr)
         distance (km)
         depth (m)
         pipe diameter (in)
         pipe thickness (in)

         Output
         ======
         pressure Drop diameter (Pa)
         pipe thickness (in)
        """

        # Calculate pipeline distance to total length of pipe
        # TODO: Replace percent_added_length with bends, drops, connections
        site_depth_km = self.site_depth_m / 1000.0 # m to km
        percent_added_length = 0.02 # 2%

        total_length = (1+percent_added_length)* self.dist_to_h2_load_km + 2 * site_depth_km

        self.output_dict['total_pipeline_length_km'] = total_length

        # Volumetric Flow Rate
        flow_rate_kg_per_sec = self.h2_flow_kg_h / 3600.0   # kg/h to kg/s
        Q = flow_rate_kg_per_sec / self.rho_h2              # m**3/s
        if verbose: print("Vol. Flow rate (m**3/sec):", Q) 

        # Pipe Dimensions
        L_m = total_length * 1000.0                     # km to m
        d_m = self.pipe_diam_in * self.in2cm / 100.0    # inch to meters 
        A = (np.pi/4) * d_m**2                          # m**2

        # Velocity and Reynolds Number 
        V = Q / A                                   # m/s
        Re = (self.rho_h2 / self.mu_h2) * V * d_m
        if verbose: print("Reynolds number:", Re)
        
        # Friction factor (assuming laminar flow)
        # TODO: Find a more accurate number for relative roughness and better Reynolds estimate 
        # if(Re < 4001.0): 
        f = 64/Re       # laminar flow
        #else: 
        #   f = 0.03        # Approx. from Moody Diagram assume Re ~10^4 and relative roughness ~0.002 
        if verbose: print("Friction factor:", f)
    
        # Design pressure option - credit undersea pressure 
        P_hydro_Pa = self.rho_h2o * self.g * self.site_depth_m # N/m**2
        press_credit = False
        if press_credit:
            P_design_Pa = self.pres_in_bar * self.bar2Pa - P_hydro_Pa
        else:
            P_design_Pa = self.pres_in_bar *self.bar2Pa
        
        P_design_psi = P_design_Pa * (self.bar2psi / self.bar2Pa )
        self.output_dict['pres_design_bar'] = P_design_psi / self.bar2psi
        
        # Approximating The Minimum Yield Stress based on SCH40 pipe
        # TODO: Expand for either more SCH's
        SYMS = P_design_psi * 1000.0 / self.pipe_sch
        
        if SYMS <=52000: 
            self.Smin = 52000
        else:
            self.Smin = SYMS

        self.output_dict['syms_sch40pipe'] = self.Smin
        # Design pressure from Barlow's Formula
        S = self.Smin * self.F * self.E * self.T * self.Hf   # Design Hoop Stress

        # Create 2D array of pressures based on diameter and thickness for a given hoop stress
        P_calc_psi = np.zeros([np.size(self.pipe_diam_in),np.size(self.pipe_thic_in)])
        for i in range(np.size(self.pipe_diam_in)):
            for j in range(np.size(self.pipe_thic_in)): 
                P_calc_psi[i,j] = 2 * self.pipe_thic_in[j] * S / self.pipe_diam_in[i]

        if verbose: print("Pressure (bar):", P_calc_psi/self.bar2psi)
        self.output_dict['pres_calc_bar'] = P_calc_psi / self.bar2psi
        
        # Find the thickness and diameter that meet the design pressure 
        # TODO: Come up with a higher fidelity method to find values in a 2D array 
        tol = 0.01 * self.output_dict['pres_design_bar'] # Within a 10% tolerance 
        d_Ind, t_Ind = np.where(np.abs(self.output_dict['pres_calc_bar'] - self.output_dict['pres_design_bar']) < tol)
        # print(d_Ind)
        if verbose: print(self.pipe_diam_in[d_Ind])
        # print(t_Ind)
        if verbose: print(self.pipe_thic_in[t_Ind])

        self.output_dict['design_diam_in'] = self.pipe_diam_in[d_Ind]
        self.output_dict['design_thic_in'] = self.pipe_thic_in[t_Ind]

            # Pressure Drop from Darcy's Equation (General Flow Equation)
            # TODO: Find outlet pressure and pressure drop
            # delP = 8.0 * ( self.rho_h2 * f * L_m * Q**2 ) / ((self.pipe_diam_in[i] * np.pi**2)**(1/5))
            #self.output_dict['pressure_drop_Pa'] = delP 
            #self.output_dict['pressure_out'] = self.pres_design_bar - delP / 100000.0 # bars 

    def pipelineCost(self):
        # Designed diameter, thickness, and total length all in meters 
        d_m = self.output_dict['design_diam_in']  * self.in2cm / 100.0 
        t_m = self.output_dict['design_thic_in']  * self.in2cm / 100.0
        L_m = self.output_dict['total_pipeline_length_km'] * 1000.0

        # Find total surface area, volume, and mass of steel required for pipeline
        pipe_cost_total = np.zeros([np.size(d_m)])

        for i in range(np.size(d_m)):
            steel_volume = np.pi * d_m[i] * t_m[i] * L_m             # m**3 
            steel_mass = steel_volume * self.rho_steel     # kg

            pipe_cost_total[i] = steel_mass * self.steel_cost_ton / 1000.0   # USD
            # TODO: Figure out what PPI to $USD per weight is 
            # fab_cost = mat_cost * 256/100 # PPI for tube manufacturing https://fred.stlouisfed.org/series/PCU33121033121002 

        if verbose: print("Design pressure Pipe Cost:",pipe_cost_total) 
        
        # Find the cost of installation based on ORBIT's example S-lay vessel
        pipe_install_hrs = self.output_dict['total_pipeline_length_km'] / 0.15
        slay_vessel_daily_rate = 0.44e6
        pipe_install_cost = ( pipe_install_hrs / 24.0 ) * slay_vessel_daily_rate

        # Cost of the hybrid substation based on ORBIT's example 
        substation_design_cost = 20e6
        substation_mass = 21000 # tons of steel for install 
        substructure_mass = 0.4 * substation_mass
        pile_mass = 8 * substructure_mass ** 0.5574
        total_substructure_mass = substructure_mass + pile_mass
        struc_steel_cost = 3000.0 # $USD/ton
        fabrication_rate = 14500 # $USD/ton
        substructure_cost = total_substructure_mass * struc_steel_cost + substation_mass * fabrication_rate
        substation_install_days = 14
        subs_vessel_daily_rate = 0.5e6
        substation_install_cost = substation_install_days * subs_vessel_daily_rate 

        # Total CapEx
        capex_pipe = pipe_cost_total + pipe_install_cost
        capex_substation = substation_design_cost + substructure_cost + substation_install_cost

        # Find the cost of construction and loading/unloading 
        self.output_dict['pipeline_capex'] = capex_pipe
        self.output_dict['substation_capex'] = capex_substation
        
        # Opex Factor based on on-land or undersea (7%) 
        # https://www.sciencedirect.com/science/article/pii/S0360319921009137?via%3Dihub 
        opex_pipe = 7/100 * capex_pipe / self.plant_life    # $US/year

        self.output_dict['pipeline_opex'] = opex_pipe

# Test sections 
if __name__ == '__main__':
    print("PipelineASME Testing section")
    in_dict = dict()
    in_dict['steel_cost_ton'] = 900.0 # $ US/ton searching for seamless FBE X52 carbon steel > $500-$1000 per ton

    in_dict['dist_to_h2_load_km'] = 80.0 
    in_dict['site_depth_m'] = 45.0
    in_dict['pipe_diam_in'] = np.linspace(12.0, 48.0, 20)
    in_dict['pipe_thic_in'] = np.linspace(0.1, 2.0, 50)

    in_dict['flow_rate_kg_hr'] = 125
    in_dict['pressure_bar'] = 100.0 
    
    out_dict = dict()

    pipeline_test = PipelineASME(in_dict,out_dict)
    pipeline_test.pipelineDesign()
    pipeline_test.pipelineCost()
 
    #print("Min. Hoop Stress SCH40 pipe:", out_dict['syms_sch40pipe'])
    print("Pipeline Length (km):", out_dict['total_pipeline_length_km'])
    print("Pipeline Design Pressure (bar):",in_dict['pressure_bar'])
    print("Pipeline Diameter: {} in, Thickness {} in".format(out_dict['design_diam_in'][0],out_dict['design_thic_in'][0]))
    print("Pipeline Cost ($US/km): ", np.min(out_dict['pipeline_capex']/out_dict['total_pipeline_length_km']))
    print("Substation Cost: ", out_dict['substation_capex'])
    print("Total H2-Export CapEx:", out_dict['substation_capex']+np.min(out_dict['pipeline_capex']))
    print("Pipeline Opex ($US/year)", np.min(out_dict['pipeline_opex']))
    
    # Contour Plot
    #fig, ax = plt.subplots()
    fig, axs = plt.subplots(figsize=(6, 4))
    #fig, axs = plt.subplots(1, 2, figsize=(9, 5), sharey=True)
    #CS1 = axs.contour(in_dict['pipe_diam_in'],in_dict['pipe_thic_in'], \
    #                  np.transpose(out_dict['pres_calc_bar']), levels = [out_dict['pres_design_bar'] ], \
    #                   colors=('k',),linestyles=('--',),linewidths=(3,))
    CS = axs.contour(in_dict['pipe_diam_in'], in_dict['pipe_thic_in'], np.transpose(out_dict['pres_calc_bar']))
    CS1 = axs.contour(in_dict['pipe_diam_in'], in_dict['pipe_thic_in'], np.transpose(out_dict['pres_calc_bar']), \
                      levels=[out_dict['pres_design_bar']], colors=('k'), linestyles=('--'),linewidths=(2))

    # def fmt(x):
    #     s = f"{x:.1f}"
    #     if s.endswith("0"):
    #         s = f"{x:.0f}"
    #     return rf"{s} " if plt.rcParams["text.usetex"] else f"{s} "
    axs.clabel(CS, CS.levels, inline=True, fontsize=10)
    axs.clabel(CS1, CS1.levels, inline=True, fontsize=10)

    axs.set_title("Pipeline Pressure Analysis")
    axs.set_xlabel("Pipe Diameter (in)")
    axs.set_ylabel("Pipe Thickness (in)")

    plt.figure(figsize=(6,4))
    plt.plot(out_dict['design_diam_in'],out_dict['pipeline_capex']/out_dict['total_pipeline_length_km'])
    plt.xlabel('Pipe Diameter (in)')
    plt.ylabel('Unit Cost ($US/km)')
    plt.title("Cost of Pipeline \n Design Pressure:{}bar".format(out_dict['pres_design_bar']))
    plt.show()
    


import pytest
from pytest import fixture

from greenheart.simulation.technologies.hydrogen.h2_storage.lined_rock_cavern.lined_rock_cavern import LinedRockCavernStorage

# Test values are based on conclusions of Papadias 2021 and are in 2019 USD
in_dict = {
    "h2_storage_kg": 1000000,
    "system_flow_rate":100000,
    "model":'papadias'}

@fixture
def lined_rock_cavern_storage():
    lined_rock_cavern_storage = LinedRockCavernStorage(in_dict)

    return lined_rock_cavern_storage


def test_init():
    lined_rock_cavern_storage = LinedRockCavernStorage(in_dict)

    assert lined_rock_cavern_storage.input_dict is not None
    assert lined_rock_cavern_storage.output_dict is not None

def test_capex_per_kg(lined_rock_cavern_storage):
    lined_rock_cavern_storage_capex_per_kg, _installed_capex, _compressor_capex = lined_rock_cavern_storage.lined_rock_cavern_capex()
    assert lined_rock_cavern_storage_capex_per_kg == pytest.approx(51.532548895265045)

def test_capex(lined_rock_cavern_storage):
    _lined_rock_cavern_storage_capex_per_kg, installed_capex, _compressor_capex = lined_rock_cavern_storage.lined_rock_cavern_capex()
    assert installed_capex == pytest.approx(51136144.673)

def test_compressor_capex(lined_rock_cavern_storage):
    _lined_rock_cavern_storage_capex_per_kg, _installed_capex, compressor_capex = lined_rock_cavern_storage.lined_rock_cavern_capex()
    assert compressor_capex == pytest.approx(9435600.2555)

def test_capex_output_dict(lined_rock_cavern_storage):
    _lined_rock_cavern_storage_capex_per_kg, _installed_capex, _compressor_capex = lined_rock_cavern_storage.lined_rock_cavern_capex()
    assert lined_rock_cavern_storage.output_dict["lined_rock_cavern_storage_capex"] == pytest.approx(51136144.673)


def test_opex(lined_rock_cavern_storage):
    _lined_rock_cavern_storage_capex_per_kg, _installed_capex, _compressor_capex = lined_rock_cavern_storage.lined_rock_cavern_capex()
    lined_rock_cavern_storage.lined_rock_cavern_opex()
    assert lined_rock_cavern_storage.output_dict["lined_rock_cavern_storage_opex"] == pytest.approx(2359700)